from tirr.packages.netvlad import pittsburgh
import pickle
import sys
import zlib
from tirr.utils.datasets import DatasetPittsTrain, DatasetPitts
import torch
from tirr.nn.models import CLIPCompressorIT, ProjectionHead
from torch.nn import functional as F
import numpy as np
import torch.nn as nn
from tirr.utils import get_embeddings
from tirr.sandbox.sandbox import get_recall

import torch
import torch.nn.functional as F

# tf writer tensorboard
from torch.utils.tensorboard import SummaryWriter

def custom_triplet_cosine_loss(anchor, positive, negative, margin=0.2):
    """
    Compute the triplet loss using cosine similarity.
    
    Parameters:
    - anchor: Tensor of shape (batch_size, features) for the anchor samples.
    - positive: Tensor of shape (batch_size, num_positives, features) for the positive samples.
    - negative: Tensor of shape (batch_size, num_negatives, features) for the negative samples.
    - margin: The margin by which the negative should be "less similar" to the anchor than the positive.
    
    Returns:
    - The average triplet loss for the batch.
    """
    # Normalize the vectors to unit length
    anchor_norm = F.normalize(anchor, p=2, dim=1)
    positive_norm = F.normalize(positive, p=2, dim=-1)
    negative_norm = F.normalize(negative, p=2, dim=-1)
    
    # Calculate the cosine similarity
    pos_similarity = torch.bmm(positive_norm, anchor_norm.unsqueeze(2)).squeeze(2)
    neg_similarity = torch.bmm(negative_norm, anchor_norm.unsqueeze(2)).squeeze(2)
    
    # Convert to distance
    pos_distance = 1 - pos_similarity
    neg_distance = 1 - neg_similarity
    
    # You could average the distances or pick the hardest samples here.
    # Here, we're taking the minimum distance among the positives and the maximum among the negatives.

    # pos_distance = pos_distance.min(dim=1)[0]  # Taking the closest positive in terms of similarity
    # neg_distance = neg_distance.max(dim=1)[0]  # Taking the furthest negative in terms of similarity
    
    # Compute the triplet loss
    losses = F.relu(pos_distance - neg_distance + margin)
    
    return losses.mean()

def eval(eval_data, model):
    (
        dbFeatImg,
        qFeatImg,
        dbFeatText,
        qFeatText,
        gt,
        numDb,
        numQ,
    ) = eval_data
    n = 3

    dbFeatComb = get_embeddings.image_text_combine(dbFeatImg, dbFeatText, batch_size=32, model=model)
    qFeatComb = get_embeddings.image_text_combine(qFeatImg, qFeatText, batch_size=32, model=model)
    recall_comb = get_recall(qFeatComb, dbFeatComb, gt, numQ, n=n)

    return recall_comb

def main():
    # Set up tensorboard writer
    writer = SummaryWriter("data/tensorboard/clip_compressor")


    pos_count = 20
    neg_count = 20
    batch_size = 32
    emb_size = 512
    dataset = DatasetPittsTrain(no_training=False, pos_count=pos_count, neg_count=neg_count)
    dataset_eval = DatasetPitts()

    eval_data = dataset_eval.get_dataset_embeddings(embed_type="clip")
    train_data = dataset.get_dataset_embeddings(embed_type="clip")


    model = CLIPCompressorIT()
    model = model.cuda()
    
    dataloder = torch.utils.data.DataLoader(
        dataset,
        batch_size=batch_size,
        shuffle=False,
        num_workers=1,
        pin_memory=True,
        drop_last=False,
    )

    optimizer = torch.optim.Adam(model.parameters(), lr=1e-4)

    # triplet_loss = ( torch.nn.TripletMarginWithDistanceLoss(distance_function=lambda x, y: 1-F.cosine_similarity(x, y)))
    triplet_loss = custom_triplet_cosine_loss
    best_recall = 0
    for epoch in range(50):
        epoch_loss = []
        for i, batch_data in enumerate(dataloder):
            optimizer.zero_grad()

            current_batch_size = batch_data[0].shape[0]

            query_img_emb, pos_img_emb, neg_img_emb, query_text_emb, pos_text_emb, neg_text_emb = batch_data
            
            img_emb_stack = torch.cat([query_img_emb, pos_img_emb, neg_img_emb], dim=1)
            text_emb_stack = torch.cat([query_text_emb, pos_text_emb, neg_text_emb], dim=1)
            
            img_emb = img_emb_stack.reshape(-1, emb_size).cuda() # verify shape
            text_emb = text_emb_stack.reshape(-1, emb_size).cuda()

            # Running model
            img_emb, text_emb = model(img_emb, text_emb)

            
            img_emb = img_emb.reshape(current_batch_size, (1+pos_count+neg_count), 50)
            text_emb = text_emb.reshape(current_batch_size, (1+pos_count+neg_count), 100)

            query_img_emb = img_emb[:, 0, :].unsqueeze(1)
            pos_img_emb = img_emb[:, 1:1+pos_count, :]
            neg_img_emb = img_emb[:, 1+pos_count:, :]

            query_text_emb = text_emb[:, 0, :].unsqueeze(1)
            pos_text_emb = text_emb[:, 1:1+pos_count, :]
            neg_text_emb = text_emb[:, 1+pos_count:, :]

            query_emb = torch.cat([query_img_emb, query_text_emb], dim=2)
            pos_emb = torch.cat([pos_img_emb, pos_text_emb], dim=2)
            neg_emb = torch.cat([neg_img_emb, neg_text_emb], dim=2)

            # loss = triplet_loss(query_emb[:,0,:], pos_emb[:,0,:], neg_emb[:,0,:])
            loss = triplet_loss(query_emb[:,0,:], pos_emb, neg_emb)
            epoch_loss.append(loss.item())
            # loss = triplet_loss(query_img_emb, pos_img_emb, neg_img_emb) # Just for sanity

            loss.backward()
            optimizer.step()

        print(f"\n\nEpoch {epoch} Loss: {np.mean(epoch_loss)}")

        # save model to data /nn
        recall_eval = eval(eval_data, model)
        recall_train = eval(train_data, model)
        if recall_eval > best_recall:
            print('Saving Best Model!')
            best_recall = recall_eval
            torch.save(model.state_dict(), "data/nn/best_model_clip_compressor.pt")

        writer.add_scalar("Loss/train", np.mean(epoch_loss), epoch)
        writer.add_scalar("Recall/eval", recall_eval, epoch)
        writer.add_scalar("Recall/train", recall_train, epoch)
        
        torch.save(model.state_dict(), "data/nn/model_clip_compressor.pt")
    
        
if __name__ == "__main__":

    main()