from tirr.sandbox.sandbox_imports import *
import yaml

class VICRegLoss(torch.nn.Module):
    def __init__(
        self,
        inv_coeff: float = 25.0,
        var_coeff: float = 15.0,
        cov_coeff: float = 1.0,
        gamma: float = 1.0,
    ):
        super().__init__()
        self.inv_coeff = inv_coeff
        self.var_coeff = var_coeff
        self.cov_coeff = cov_coeff
        self.gamma = gamma

    def forward(self, x: torch.Tensor, y: torch.Tensor):
        """Computes the VICReg loss.

        ---
        Args:
            x: Features map.
                Shape of [batch_size, representation_size].
            y: Features map.
                Shape of [batch_size, representation_size].

        ---
        Returns:
            The VICReg loss.
                Dictionary where values are of shape of [1,].
        """
        metrics = dict()
        metrics["inv-loss"] = self.inv_coeff * self.representation_loss(x, y)
        metrics["var-loss"] = (
            self.var_coeff
            * (
                self.variance_loss(x, self.gamma)
                + self.variance_loss(y, self.gamma)
            )
            / 2
        )
        metrics["cov-loss"] = (
            self.cov_coeff
            * (self.covariance_loss(x) + self.covariance_loss(y))
            / 2
        )
        metrics["loss"] = sum(metrics.values())
        return metrics

    @staticmethod
    def representation_loss(x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        """Computes the representation loss.
        Force the representations of the same object to be similar.

        ---
        Args:
            x: Features map.
                Shape of [batch_size, representation_size].
            y: Features map.
                Shape of [batch_size, representation_size].

        ---
        Returns:
            The representation loss.
                Shape of [1,].
        """
        return F.mse_loss(x, y)

    @staticmethod
    def variance_loss(x: torch.Tensor, gamma: float) -> torch.Tensor:
        """Computes the variance loss.
        Push the representations across the batch
        to be different between each other.
        Avoid the model to collapse to a single point.

        The gamma parameter is used as a threshold so that
        the model is no longer penalized if its std is above
        that threshold.

        ---
        Args:
            x: Features map.
                Shape of [batch_size, representation_size].

        ---
        Returns:
            The variance loss.
                Shape of [1,].
        """
        x = x - x.mean(dim=0)
        std = x.std(dim=0)
        var_loss = F.relu(gamma - std).mean()
        return var_loss

    @staticmethod
    def variance_loss_strict(x: torch.Tensor, gamma: float) -> torch.Tensor:
        """Computes the variance loss.
        Push the representations across the batch
        to be different between each other.
        Avoid the model to collapse to a single point.

        The gamma parameter is the desired std of the representations.

        ---
        Args:
            x: Features map.
                Shape of [batch_size, representation_size].

        ---
        Returns:
            The variance loss.
                Shape of [1,].
        """
        x = x - x.mean(dim=0)
        std = x.std(dim=0)
        var_loss = F.mse_loss(std, torch.ones_like(std) * gamma)
        return var_loss

    @staticmethod
    def covariance_loss(x: torch.Tensor) -> torch.Tensor:
        """Computes the covariance loss.
        Decorrelates the embeddings' dimensions, which pushes
        the model to capture more information per dimension.

        ---
        Args:
            x: Features map.
                Shape of [batch_size, representation_size].

        ---
        Returns:
            The covariance loss.
                Shape of [1,].
        """
        x = x - x.mean(dim=0)
        cov = (x.T @ x) / (x.shape[0] - 1)
        cov_loss = cov.fill_diagonal_(0.0).pow(2).sum() / x.shape[1]
        return cov_loss


APPROACHES = {
    "VICReg": False,
    "AESVC": False,
    "VAE": False,
    "AE": False,
}

def infer_ae(
    embedding="dino",
    dataset_name="inshop",
    metric_name="map",
    distance="cosine",
):
    dataset_mapping = get_dataset_mapping()

    # creating an instance of the required dataset class
    dataset = dataset_mapping[dataset_name](embed_type=embedding
    )

    comp_range = [2**i for i in range(2, int(np.log2(dataset.final_dim_size)) + 1)] + [dataset.final_dim_size]
    
    metrics_mapping = get_metrics_mapping()
    metric_function = metrics_mapping[metric_name]

    # create a yaml file that stores results for all approaches for all embedding dimensions
   

    results = {key: {} for key in APPROACHES.keys()}
    results['PCA'] = {}

    


    for bottle_neck_size in comp_range:
        print(f"Embedding Dimension: {bottle_neck_size}")
        for approach_name in list(APPROACHES.keys()):
            if approach_name == "VAE":
                model = VariationalAE(
                    emb_size=dataset.final_dim_size, bottle_neck_size=bottle_neck_size
                )
            elif approach_name == "VICReg":
                model = AutoEncoderVicReg(
                    emb_size=dataset.final_dim_size, bottle_neck_size=bottle_neck_size
                )
            else:
                model = AutoEncoder(
                    emb_size=dataset.final_dim_size, bottle_neck_size=bottle_neck_size
                )
            
            model.load_state_dict(
                torch.load(
                    f"data/models/{approach_name}_{dataset_name}_{embedding}_{bottle_neck_size}.pth"
                )
            )
            model = model.cuda()

            qFeatImgProj = (
                model.encode(torch.tensor(dataset.qFeatImg).float().cuda())
                .detach()
                .cpu()
                .numpy()
            )
            dbFeatImgProj = (
                model.encode(torch.tensor(dataset.dbFeatImg).float().cuda())
                .detach()
                .cpu()
                .numpy()
            )

            recall_proj = metric_function(
                qFeatImgProj.copy(),
                dbFeatImgProj.copy(),
                dataset.gt,
                dataset.numQ,
                k=4,
                distance=distance,
            )
            print(f"Recall Image {approach_name}: {recall_proj}")


            results[approach_name][bottle_neck_size] = float(recall_proj)

        recall_img = metric_function(
                dataset.qFeatImg_PCA[:, :bottle_neck_size].copy(),
                dataset.dbFeatImg_PCA[:, :bottle_neck_size].copy(),
                dataset.gt,
                dataset.numQ,
                k=4,
                distance=distance,
            )
        results['PCA'][bottle_neck_size] = float(recall_img)
        print(f"Recall Image PCA: {recall_img}")

    with open(f"data/results/iclr_{dataset_name}_{embedding}_{metric_name}.yaml", "w") as file:
        yaml.dump(results, file, default_flow_style=False)

    
        

def train_ae(
    embedding="dino",
    dataset_name="inshop",
    metric_name="map",
    distance="cosine",
    bottle_neck_size=64,
    approach_name="AESVC",
):
    # Hyperparameters
    batch_size = 64

    num_epochs = 3
    lr = 1e-3

    dataset_mapping = get_dataset_mapping()

    APPROACHES[approach_name] = True

    # creating an instance of the required dataset class
    dataset = dataset_mapping[dataset_name](embed_type=embedding
    )

    metrics_mapping = get_metrics_mapping()

    metric_function = metrics_mapping[metric_name]

    if APPROACHES["VAE"]:
        model = VariationalAE(
            emb_size=dataset.final_dim_size, bottle_neck_size=bottle_neck_size
        )
    elif APPROACHES["VICReg"]:
        model = AutoEncoderVicReg(
            emb_size=dataset.final_dim_size, bottle_neck_size=bottle_neck_size
        )
    else:
        model = AutoEncoder(
            emb_size=dataset.final_dim_size, bottle_neck_size=bottle_neck_size
        )
    model = model.cuda()

    dataloder = torch.utils.data.DataLoader(
        dataset,
        batch_size=batch_size,
        shuffle=True,
        num_workers=1,
        pin_memory=True,
        drop_last=False,
    )

    optimizer = torch.optim.AdamW(model.parameters(), lr=lr)

    final_dim = dataset.final_dim_size

    best_recall = 0

    mse_loss = torch.nn.MSELoss()
    ce_loss = torch.nn.CrossEntropyLoss()

    criterion_ = torch.nn.CosineSimilarity()
    criterion_cosine_loss = lambda x, y: -criterion_(x, y).mean()

    vicreg = VICRegLoss()

    best_test_recall = 0

    recall_default = metric_function(
        dataset.qFeatImg.copy(),
        dataset.dbFeatImg.copy(),
        dataset.gt,
        dataset.numQ,
        k=4,
        distance=distance,
    )

    print(f"Recall Default: {recall_default}")

    for epoch in range(num_epochs):
        epoch_loss = []
        for i, batch_data in enumerate(tqdm(dataloder)):
            optimizer.zero_grad()

            img_emb1, sim_values = batch_data
            org_shape = img_emb1.shape

            img_emb1 = img_emb1.reshape(-1, final_dim)

            img_emb1 = img_emb1.cuda()

            loss = 0

            if APPROACHES["VAE"]:
                decoded, encoded, mu, logvar = model(img_emb1)
            else:
                encoded, decoded = model(img_emb1)

            # loss += criterion_cosine_loss(img_emb1, decoded)
            loss += mse_loss(img_emb1, decoded)

            ##### OUR LOSS
            # loss = vicreg.representation_loss(decoded, img_emb1)
            # # loss += vicreg.variance_loss(encoded, 1)
            if APPROACHES["AESVC"]:
                loss += 10 * vicreg.variance_loss_strict(encoded, 1)
                loss += 10 * vicreg.covariance_loss(encoded)

            #### Default VicReg Loss
            if APPROACHES["VICReg"]:
                loss += vicreg.variance_loss(decoded, 1)
                loss += vicreg.covariance_loss(decoded)

            if APPROACHES["VAE"]:
                KLD = -0.5 * torch.sum(1 + logvar - mu.pow(2) - logvar.exp())
                KLD /= batch_size * final_dim
                KLD_weight = 0.01
                loss += KLD_weight * KLD

            epoch_loss.append(loss.item())

            loss.backward()
            optimizer.step()

        print(f"Epoch: {epoch}, Loss: {np.mean(epoch_loss)}")

        qFeatImgProj = (
            model.encode(torch.tensor(dataset.qFeatImg).float().cuda())
            .detach()
            .cpu()
            .numpy()
        )
        dbFeatImgProj = (
            model.encode(torch.tensor(dataset.dbFeatImg).float().cuda())
            .detach()
            .cpu()
            .numpy()
        )
        if epoch % 1 == 0:

            recall_proj = metric_function(
                qFeatImgProj.copy(),
                dbFeatImgProj.copy(),
                dataset.gt,
                dataset.numQ,
                k=4,
                distance=distance,
            )
            print(f"Recall Image Proj: {recall_proj}")

            if recall_proj > best_recall:
                best_recall = recall_proj
                torch.save(
                    model.state_dict(),
                    f"data/models/{approach_name}_{dataset_name}_{embedding}_{bottle_neck_size}.pth",
                )
            print('Saved model to ', f"/store/omama/TIRR/tirr/data/models/{approach_name}_{dataset_name}_{embedding}_{bottle_neck_size}.pth")
            print(f"Best recall: {best_recall}")

            recall_img = metric_function(
                dataset.qFeatImg_PCA[:, :bottle_neck_size].copy(),
                dataset.dbFeatImg_PCA[:, :bottle_neck_size].copy(),
                dataset.gt,
                dataset.numQ,
                k=4,
                distance=distance,
            )
            print(f"Recall Image PCA : {recall_img}")


def main():
    dataset_name = "tokyo_val"
    embed_type = "dinos8"

    approach_name = "VICReg"

    # create dataset class
    dataset_mapping = get_dataset_mapping()
    dataset = dataset_mapping[dataset_name](embed_type=embed_type)
    # comp_range = [2**i for i in range(2, int(np.log2(dataset.final_dim_size)) + 1)] + [dataset.final_dim_size]
    comp_range = [256,384]

    for bottle_neck in comp_range:
        train_ae(
            embedding=embed_type,
            dataset_name=dataset_name,
            metric_name="map",
            distance="cosine",
            bottle_neck_size=bottle_neck,
            approach_name=approach_name,
        )


if __name__ == "__main__":
    # main()
    infer_ae(
        embedding="dinos8",
        dataset_name="tokyo_val",
        metric_name="map",
        distance="cosine",
    )


"""
0.5381: VC-VAE

"""