from tirr.sandbox.sandbox_imports import *


def weighted_corr_sim(
    X: np.ndarray, Y: np.ndarray, corr: np.ndarray, dim: int = 150
) -> np.ndarray:
    """
    Compute the weighted correlation similarity
    :param X: data 1. shape: (N, feats)
    :param Y: data 2. shape: (N, feats)
    :param corr: correlation matrix. shape: (feats, )
    :return: similarity matrix between X and Y. shape: (N, )
    """

    assert (
        X.shape == Y.shape
    ), f"X and Y should have the same number of shape, but got {X.shape} and {Y.shape}"
    # select the first dim dimensions
    X, Y, corr = X[:, :dim], Y[:, :dim], corr[:dim]
    # normalize X and Y with L2 norm
    X = X / np.linalg.norm(X, axis=1, keepdims=True)
    Y = Y / np.linalg.norm(Y, axis=1, keepdims=True)
    # compute the similarity scores
    sim = np.zeros(X.shape[0])
    for ii in range(X.shape[0]):
        sim[ii] = corr * X[ii] @ Y[ii]
    return sim


def get_pca_reprojected_emb(pca, dbFeatImg, qFeatImg, quantization_level=None):
    dbFeatImgPca = pca.fit_transform(dbFeatImg)
    qFeatImgPca = pca.transform(qFeatImg)

    if quantization_level is not None:
        print("Quantizing the embeddings")
        dbFeatImgPca = quantize_tensor(
            dbFeatImgPca, num_bits=quantization_level
        )
        dbFeatImgPca = dequantize_tensor(dbFeatImgPca)
        qFeatImgPca = quantize_tensor(qFeatImgPca, num_bits=quantization_level)
        qFeatImgPca = dequantize_tensor(qFeatImgPca)

    dbFeatImgHat = pca.inverse_transform(dbFeatImgPca)
    qFeatImgHat = pca.inverse_transform(qFeatImgPca)

    return dbFeatImgHat, qFeatImgHat


def get_cpca_reprojected_emb(pca, dbFeatImg, qFeatImg, dbFeatText):
    dbFeatImgPca = pca.fit_transform(dbFeatImg, dbFeatText)
    qFeatImgPca = pca.transform_fg(qFeatImg)

    dbFeatImgHat = pca.inverse_transform_fg(dbFeatImgPca)
    qFeatImgHat = pca.inverse_transform_fg(qFeatImgPca)

    return dbFeatImgHat, qFeatImgHat


def reduced_embeddings_exp(n=3):
    dataset = DatasetKITTI()
    (
        dbFeatImg,
        qFeatImg,
        dbFeatText,
        qFeatText,
        gt,
        numDb,
        numQ,
    ) = dataset.get_dataset_embeddings("clip")

    logger.info("Getting Predictions")
    # Predictions for images
    recall_img = get_recall(qFeatImg, dbFeatImg, gt, numQ, n=n)
    print(f"Recall@{n} for images: = {recall_img}")

    # Predictions for text
    recall_text = get_recall(qFeatText, dbFeatText, gt, numQ, n=n)
    print(f"Recall@{n} for text: = {recall_text}")

    comp_range = [2, 5, 10, 50, 60, 70, 80, 90, 100]
    results = {
        "pca": [],
        "pca_text": [],
        "cpca_text": [],
        "comp_range": comp_range,
        "recall_img": recall_img,
        "recall_text": recall_text,
    }

    for pca_comp in tqdm(comp_range):
        print("Embedding Dimension:", pca_comp)
        imgPca = PCA(n_components=pca_comp)
        dbFeatImgPca = imgPca.fit_transform(dbFeatImg)
        qFeatImgPca = imgPca.transform(qFeatImg)

        recall = get_recall(qFeatImgPca, dbFeatImgPca, gt, numQ, n=n)
        print(f"Recall@{n} for images with PCA: = {recall}")
        results["pca"].append(recall)

        ##### predictions for image pca with text transofrm
        pca = PCA(n_components=pca_comp)
        dbFeatImgHat, qFeatImgHat = get_pca_reprojected_emb(
            pca, dbFeatImg, qFeatImg
        )

        dbFeatImgHatText = dbFeatImgHat + dbFeatText
        qFeatImgHatText = qFeatImgHat + qFeatText

        recall = get_recall(qFeatImgHatText, dbFeatImgHatText, gt, numQ, n=n)
        print(f"Recall@{n} for images with PCA and text transform: = {recall}")
        results["pca_text"].append(recall)

        ##### predictions for customCpca with text transofrm
        alpha = 0.1
        pca = CustomCPCA(n_components=pca_comp, alpha=alpha)
        dbFeatImgHat, qFeatImgHat = get_cpca_reprojected_emb(
            pca, dbFeatImg, qFeatImg, dbFeatText
        )

        dbFeatImgHatText = dbFeatImgHat + dbFeatText
        qFeatImgHatText = qFeatImgHat + qFeatText

        recall = get_recall(qFeatImgHatText, dbFeatImgHatText, gt, numQ, n=n)
        print(
            f"Recall@{n} for images with Custom CPCA and text transform: = {recall}"
        )
        results["cpca_text"].append(recall)

    # save results pickle
    with open(f"data/pkl/reduced_emb_recall@{n}.pkl", "wb") as f:
        pickle.dump(results, f)


def weighted_sum_exp(n=3):
    dataset = DatasetPittsTrain()
    (
        dbFeatImg,
        qFeatImg,
        dbFeatText,
        qFeatText,
        gt,
        numDb,
        numQ,
    ) = dataset.get_dataset_embeddings(embed_type="clip")

    logger.info("Getting Predictions")
    # Predictions for images
    recall_img = get_recall(qFeatImg, dbFeatImg, gt, numQ, n=n)
    print(f"Recall@{n} for images: = {recall_img}")

    # Predictions for text
    recall_text = get_recall(qFeatText, dbFeatText, gt, numQ, n=n)
    print(f"Recall@{n} for text: = {recall_text}")

    # comp_range = [2, 5, 10, 20, 30, 40, 60, 80, 100, 140, 160, 200, 250, 300, 350, 400, 500, 600, 700]
    comp_range = [10, 40, 50, 60, 70, 100, 120, 170, 200, 250, 300, 400]
    results = {
        "pca": [],
        "pca_text": [],
        "comp_range": comp_range,
        "recall_img": recall_img,
        "recall_text": recall_text,
    }

    dbFeatImg = np.sign(dbFeatImg)
    qFeatImg = np.sign(qFeatImg)

    for pca_comp in tqdm(comp_range):
        print("Embedding Dimension:", pca_comp)
        imgPca = PCA(n_components=pca_comp)
        # _ = imgPca.fit_transform(np.concatenate([dbFeatImg,dbFeatImgPits, dbFeatImgPits, dbFeatImgPits]))
        _ = imgPca.fit_transform(dbFeatImg)
        dbFeatImgPca = imgPca.transform(dbFeatImg)
        qFeatImgPca = imgPca.transform(qFeatImg)

        # normalize
        recall = get_recall(qFeatImgPca, dbFeatImgPca, gt, numQ, n=n)
        print(f"Recall@{n} for images with PCA: = {recall}")

        results["pca"].append(recall)

        # predictions for image pca with text transofrm
        pca = PCA(n_components=pca_comp)
        dbFeatImgHat, qFeatImgHat = get_pca_reprojected_emb(
            pca, dbFeatImg, qFeatImg
        )

        w = 0.5
        # if pca_comp > 8: w = 0.1
        # if pca_comp < 10: w = 0.1
        # elif pca_comp < 30: w = 0.5
        # elif pca_comp < 70: w = 0.2
        # elif pca_comp <= 100: w = 0.2
        # if pca_comp >= 100: w = 0.1

        dbFeatImgHatText = np.hstack((dbFeatImgHat, w * dbFeatText))
        qFeatImgHatText = np.hstack((qFeatImgHat, w * qFeatText))

        recall = get_recall(qFeatImgHatText, dbFeatImgHatText, gt, numQ, n=n)
        print(f"Recall@{n} for images with PCA and text transform: = {recall}")
        results["pca_text"].append(recall)

    # save results pickle
    with open(f"data/pkl/reduced_emb_recall@{n}.pkl", "wb") as f:
        pickle.dump(results, f)


def quantization_exp(n=3, quantization_level=4):
    similarity = torch.nn.CosineSimilarity(dim=1)
    dataset = DatasetPittsTrain()
    (
        dbFeatImg,
        qFeatImg,
        dbFeatText,
        qFeatText,
        gt,
        numDb,
        numQ,
    ) = dataset.get_dataset_embeddings(embed_type="vit")

    logger.info("Getting Predictions")
    # Predictions for images
    recall_img = get_recall(qFeatImg, dbFeatImg, gt, numQ, n=n)
    print(f"Recall@{n} for images: = {recall_img}")

    # Predictions for text
    recall_text = get_recall(qFeatText, dbFeatText, gt, numQ, n=n)
    print(f"Recall@{n} for text: = {recall_text}")

    # # # binary quantization for img
    # dbFeatImg = binary_quantization(dbFeatImg)
    # qFeatImg = binary_quantization(qFeatImg)

    # recall_img = get_recall(qFeatImg, dbFeatImg, gt, numQ, n=n)
    # print(f"Recall@{n} for images with binary quantization: = {recall_img}")

    comp_range = [10, 40, 50, 60, 70, 100, 120, 170, 200, 250, 300, 400]
    results = {
        "pca": [],
        "pca_text": [],
        "comp_range": comp_range,
        "recall_img": recall_img,
        "recall_text": recall_text,
    }

    for pca_comp in tqdm(comp_range):
        print("Embedding Dimension:", pca_comp)
        imgPca = PCA(n_components=pca_comp)
        # _ = imgPca.fit_transform(np.concatenate([dbFeatImg,dbFeatImgPits, dbFeatImgPits, dbFeatImgPits]))
        _ = imgPca.fit_transform(dbFeatImg)
        dbFeatImgPca = imgPca.transform(dbFeatImg)
        qFeatImgPca = imgPca.transform(qFeatImg)

        # dbFeatImgPca = binary_quantization(dbFeatImgPca)
        # qFeatImgPca = binary_quantization(qFeatImgPca)

        dbFeatImgPca = quantize_tensor(
            dbFeatImgPca, num_bits=quantization_level
        )
        dbFeatImgPca = dequantize_tensor(dbFeatImgPca)
        qFeatImgPca = quantize_tensor(qFeatImgPca, num_bits=quantization_level)
        qFeatImgPca = dequantize_tensor(qFeatImgPca)

        # normalize
        recall = get_recall(qFeatImgPca, dbFeatImgPca, gt, numQ, n=n)
        print(f"Recall@{n} for images with PCA: = {recall}")

        results["pca"].append(recall)

        # predictions for image pca with text transofrm
        pca = PCA(n_components=pca_comp)
        dbFeatImgHat, qFeatImgHat = get_pca_reprojected_emb(
            pca, dbFeatImg, qFeatImg, quantization_level
        )

        w = 0.5

        dbFeatImgHatText = np.hstack((dbFeatImgHat, w * dbFeatText))
        qFeatImgHatText = np.hstack((qFeatImgHat, w * qFeatText))

        recall = get_recall(qFeatImgHatText, dbFeatImgHatText, gt, numQ, n=n)
        print(f"Recall@{n} for images with PCA and text transform: = {recall}")
        results["pca_text"].append(recall)

    # save results pickle
    with open(
        f"data/pkl/quantization@{quantization_level}_recall@{n}.pkl", "wb"
    ) as f:
        pickle.dump(results, f)


def dcca(n=3):
    # dataset = DatasetKITTI()
    dataset = DatasetInShop()
    (
        dbFeatImg,
        qFeatImg,
        dbFeatText,
        qFeatText,
        gt,
        numDb,
        numQ,
    ) = dataset.get_dataset_embeddings(embed_type="dinov2_small")

    (
        dbFeatImg2,
        qFeatImg2,
        dbFeatText2,
        qFeatText2,
        gt2,
        numDb2,
        numQ2,
    ) = dataset.get_dataset_embeddings(embed_type="dinob8")

    # Other hyperparameters
    q_in_db = False

    logger.info("Getting Predictions")
    # Predictions for images
    recall_img = get_recall(qFeatImg, dbFeatImg, gt, numQ, n=n, q_in_db=q_in_db)
    print(f"Recall@{n} for images: = {recall_img}")

    comp_range = [
        16,
        20,
        30,
        40,
        60,
        80,
        100,
        200,
        300,
        500,
        # 700,
    ]

    results = {
        "pca": [],
        "pca_text": [],
        "cca_text": [],
        "comp_range": comp_range,
        "recall_img": recall_img,
        "recall_text": 0,
    }

    x = dbFeatImg - np.mean(dbFeatImg, axis=0)
    y = dbFeatImg2 - np.mean(dbFeatImg2, axis=0)
    # random shuffle x and y, select all
    Xs = [x, y]
    features1 = Xs[0].shape[1]
    features2 = Xs[1].shape[1]

    num_epochs = 300

    cca_dims = 100
    layers1 = [300]  # nodes in each hidden layer
    layers2 = [300]

    dcca = DCCA(
        input_size1=features1,
        input_size2=features2,
        n_components=cca_dims,
        layer_sizes1=layers1,
        layer_sizes2=layers2,
        epoch_num=num_epochs,
        device="cuda",
        use_all_singular_values=False,
        learning_rate=3 * 1e-4,
        batch_size=3000,
    )
    dcca.fit(Xs)

    dbFeatImgCCA, dbFeatImg2CCA = dcca.transform(Xs)

    x = qFeatImg - np.mean(qFeatImg, axis=0)
    y = qFeatImg2 - np.mean(qFeatImg2, axis=0)
    Xsq = [x, y]
    qFeatImgCCA, qFeatImg2CCA = dcca.transform(Xsq)

    recall = get_recall(
        qFeatImgCCA, dbFeatImgCCA, gt, numQ, n=n, q_in_db=q_in_db
    )
    print(f"Recall@{n} for images with DCCA: = {recall}")

    imgCCA = CCA(latent_dims=cca_dims)
    x = dbFeatImg - np.mean(dbFeatImg, axis=0)
    y = dbFeatImg2 - np.mean(dbFeatImg2, axis=0)
    imgCCA.fit((x, y))

    x = dbFeatImg - np.mean(dbFeatImg, axis=0)
    y = dbFeatImg2 - np.mean(dbFeatImg2, axis=0)
    dbFeatImgCCA, _ = imgCCA.transform(
        (x, y * 0)
    )  # multiplying by zero since we are only interested in the first part

    x = qFeatImg - np.mean(qFeatImg, axis=0)
    y = qFeatImg2 - np.mean(qFeatImg2, axis=0)
    qFeatImgCCA, _ = imgCCA.transform((x, y * 0))
    qFeatImgCCA = qFeatImgCCA

    recall = get_recall(
        qFeatImgCCA, dbFeatImgCCA, gt, numQ, n=n, q_in_db=q_in_db
    )
    print(f"Recall@{n} for images with CCA: = {recall}")

    breakpoint()


def cca_zoo(n=3, student_embedding="dino", teacher_embedding="dinov2"):
    dataset = DatasetInShop()
    (
        dbFeatImg,
        qFeatImg,
        dbFeatText,
        qFeatText,
        gt,
        numDb,
        numQ,
    ) = dataset.get_dataset_embeddings(embed_type=student_embedding)

    (
        dbFeatImg2,
        qFeatImg2,
        dbFeatText2,
        qFeatText2,
        gt2,
        numDb2,
        numQ2,
    ) = dataset.get_dataset_embeddings(embed_type=teacher_embedding)

    smaller_dim = min(dbFeatImg.shape[1], dbFeatImg2.shape[1])

    comp_range = list(range(20, smaller_dim + 1, 30))

    # Other hyperparameters
    q_in_db = False

    logger.info("Getting Predictions")
    # Predictions for student
    recall_img_student = get_recall(
        qFeatImg.copy(), dbFeatImg.copy(), gt, numQ, n=n, q_in_db=q_in_db
    )
    print(
        f"Recall@{n} for student embedding {student_embedding}: = {recall_img_student}"
    )

    # predictions for teacher
    recall_img_teacher = get_recall(
        qFeatImg2.copy(), dbFeatImg2.copy(), gt2, numQ2, n=n, q_in_db=q_in_db
    )
    print(
        f"Recall@{n} for teacher embedding {teacher_embedding}: = {recall_img_teacher}"
    )

    results = {
        f"pca_{student_embedding}": [],
        f"pca_{teacher_embedding}": [],
        f"{teacher_embedding} guided {student_embedding} (CCA)": [],
        "comp_range": comp_range,
        "student_embedding": recall_img_student,
        "teacher_embedding": recall_img_teacher,
    }

    imgCCA = CCA(latent_dims=comp_range[-1])
    # imgCCA = KCCA(latent_dims=comp_range[-1], kernel=["rbf", "rbf"], c=[0.9, 0.9], gamma=[1, 1])
    x = dbFeatImg - np.mean(dbFeatImg, axis=0)
    y = dbFeatImg2 - np.mean(dbFeatImg2, axis=0)
    x, y = imgCCA.fit_transform((x, y))
    corr_unalign = np.diag(x.T @ y) / x.shape[0]

    studentPca = PCA(n_components=comp_range[-1])
    _ = studentPca.fit_transform(dbFeatImg)

    TeacherPca = PCA(n_components=comp_range[-1])
    _ = TeacherPca.fit_transform(dbFeatImg2)

    cca_param_num = imgCCA.weights[0].shape[0] * imgCCA.weights[0].shape[1]
    print(f"Number of parameters in CCA (in M): {cca_param_num / 1e6 } ")

    for pca_comp in tqdm(comp_range):
        print("Embedding Dimension:", pca_comp)

        ##### Student Embeddings

        dbFeatImgPca = studentPca.transform(dbFeatImg)[:, :pca_comp]
        qFeatImgPca = studentPca.transform(qFeatImg)[:, :pca_comp]

        # mean center and unit variance dbFeatImgPCA and qFeatImgPca
        dbFeatImgPca = (dbFeatImgPca - np.mean(dbFeatImgPca, axis=0)) / np.std(
            dbFeatImgPca, axis=0
        )
        qFeatImgPca = (qFeatImgPca - np.mean(qFeatImgPca, axis=0)) / np.std(
            qFeatImgPca, axis=0
        )

        recall = get_recall(
            qFeatImgPca.copy(),
            dbFeatImgPca.copy(),
            gt,
            numQ,
            n=n,
            q_in_db=q_in_db,
        )
        print(f"Recall@{n} for {student_embedding} images with PCA: = {recall}")

        results[f"pca_{student_embedding}"].append(recall)

        ##### Teacher Embeddings

        dbFeatImgPca = TeacherPca.transform(dbFeatImg2)[:, :pca_comp]
        qFeatImgPca = TeacherPca.transform(qFeatImg2)[:, :pca_comp]

        # mean center and unit variance dbFeatImgPCA and qFeatImgPca
        dbFeatImgPca = (dbFeatImgPca - np.mean(dbFeatImgPca, axis=0)) / np.std(
            dbFeatImgPca, axis=0
        )
        qFeatImgPca = (qFeatImgPca - np.mean(qFeatImgPca, axis=0)) / np.std(
            qFeatImgPca, axis=0
        )

        recall = get_recall(
            qFeatImgPca.copy(),
            dbFeatImgPca.copy(),
            gt,
            numQ,
            n=n,
            q_in_db=q_in_db,
        )
        print(f"Recall@{n} for {teacher_embedding} images with PCA: = {recall}")

        results[f"pca_{teacher_embedding}"].append(recall)

        ##### Teacher Guided Student [CCA]
        x = dbFeatImg - np.mean(dbFeatImg, axis=0)
        y = dbFeatImg2 - np.mean(dbFeatImg2, axis=0)
        dbFeatImgCCA, dbFeatTextCCA = imgCCA.transform(
            (x, y * 0)
        )  # multiplying by zero since we are only interested in the first part
        dbFeatImgCCA = dbFeatImgCCA[:, :pca_comp]

        x = qFeatImg - np.mean(qFeatImg, axis=0)
        y = qFeatImg2 - np.mean(qFeatImg2, axis=0)
        qFeatImgCCA, qFeatTextCCA = imgCCA.transform((x, y * 0))
        qFeatImgCCA = qFeatImgCCA[:, :pca_comp]

        recall = get_recall(
            qFeatImgCCA.copy(),
            dbFeatImgCCA.copy(),
            gt,
            numQ,
            n=n,
            q_in_db=q_in_db,
        )
        print(
            f"Recall@{n} for {teacher_embedding} guided {student_embedding} [CCA]: = {recall}"
        )

        results[f"{teacher_embedding} guided {student_embedding} (CCA)"].append(
            recall
        )

    # save results pickle
    with open(
        f"data/pkl/CCA_precision@{n}_{teacher_embedding}_guided_{student_embedding}.pkl",
        "wb",
    ) as f:
        pickle.dump(results, f)


def cca_vs_pca_gaussian():
    set_plot_properties()
    number_of_samples = 20000

    weight_kg = np.random.normal(0, 1, number_of_samples) * 20
    weigh_lbs = weight_kg * 2.20462 + np.random.normal(0, 1, number_of_samples)
    noise1 = np.random.normal(0, 1, number_of_samples) * 20
    noise2 = (
        np.random.normal(0, 1, number_of_samples) * 20 * 2.20462
    )  # keeping noise in the same scale as weight_lbs

    # stack weights_kg and noise1
    X_nn1 = np.vstack((weight_kg, noise1)).T  # data captured by nn1
    X_nn2 = np.vstack((noise2, weigh_lbs)).T  # data captured by nn2

    # plot X_nn1
    plt.scatter(X_nn1[:, 0], X_nn1[:, 1], s=0.1, c="g")
    plt.xlabel("Weight (kg)")
    plt.ylabel("Noise")
    plt.title("Data captured by NN1")
    plt.savefig("data/plots/cca_vs_pca_nn1.png")
    plt.clf()

    # plot X_nn2
    plt.scatter(X_nn2[:, 0], X_nn2[:, 1], s=0.1, c="y")
    plt.xlabel("Noise")
    plt.ylabel("Weight (lbs)")
    plt.title("Data captured by NN2")
    plt.savefig("data/plots/cca_vs_pca_nn2.png")
    plt.clf()

    # PCA of X_nn1
    pca = PCA(n_components=1)
    X_nn1_pca = pca.fit_transform(X_nn1)
    pca_topvector_nn1 = pca.components_[0]
    print(f"Top Informative Vector (PCA): {pca_topvector_nn1}")

    # CCA of X_nn1 and X_nn2
    cca = skCCA(n_components=1)
    cca.fit(X_nn1, X_nn2)
    X_nn1_cca, _ = cca.transform(X_nn1, X_nn2)
    cca_topvector_nn1 = np.array([cca.x_weights_[0][0], cca.x_weights_[1][0]])
    print(f"Top Informative Vector (CCA): {cca_topvector_nn1}")

    # plot X_nn1 with eigen vector
    plt.scatter(X_nn1[:, 0], X_nn1[:, 1], s=0.1, c="g")
    plt.xlabel("Weight (kg)")
    plt.ylabel("Noise")
    plt.title("Data captured by NN1")
    plt.quiver(
        *np.mean(X_nn1, axis=0),
        *pca_topvector_nn1,
        scale=4,
        color="k",
        label="Top Informative Vector (PCA)",
    )
    plt.quiver(
        *np.mean(X_nn1, axis=0),
        *cca_topvector_nn1,
        scale=4,
        color="r",
        label="Top Informative Vector (CCA)",
    )
    plt.legend()
    plt.savefig("data/plots/cca_vs_pca_nn1_with_eigen_vector.png")
    plt.clf()

    # PCA for X_nn2
    pca = PCA(n_components=1)
    X_nn2_pca = pca.fit_transform(X_nn2)
    pca_topvector_nn2 = pca.components_[0]
    print(f"Top Informative Vector (PCA): {pca_topvector_nn2}")

    # CCa for X_nn2 and X_nn1
    cca = skCCA(n_components=1)
    cca.fit(X_nn2, X_nn1)
    X_nn2_cca, _ = cca.transform(X_nn2, X_nn1)
    breakpoint()
    cca_topvector_nn2 = np.array([cca.x_weights_[0][0], cca.x_weights_[1][0]])
    print(f"Top Informative Vector (CCA): {cca_topvector_nn2}")

    # plot X_nn2 with eigen vector
    plt.scatter(X_nn2[:, 0], X_nn2[:, 1], s=0.1, c="y")
    plt.xlabel("Noise")
    plt.ylabel("Weight (lbs)")
    plt.title("Data captured by NN2")
    plt.quiver(
        *np.mean(X_nn2, axis=0),
        *pca_topvector_nn2,
        scale=4,
        color="k",
        label="Top Informative Vector (PCA)",
    )
    plt.quiver(
        *np.mean(X_nn2, axis=0),
        *cca_topvector_nn2,
        scale=4,
        color="r",
        label="Top Informative Vector (CCA)",
    )
    plt.legend()
    plt.savefig("data/plots/cca_vs_pca_nn2_with_eigen_vector.png")


def waterbird_test():
    dataset = DatasetWaterBird()
    emb_dir = "/nas/tirr/waterbird/data/"

    emb1_file = emb_dir + "waterbird_img_emb_train_test_clip.pkl"
    emb2_file = emb_dir + "waterbird_img_emb_train_test_dino.pkl"

    with open(emb1_file, "rb") as f:
        emb1 = pickle.load(f)

    with open(emb2_file, "rb") as f:
        emb2 = pickle.load(f)

    imgCCA = CCA(latent_dims=300)
    breakpoint()


def cca_tsne(n=20, student_embedding="dino", teacher_embedding="dinov2"):
    dataset = DatasetInShop()  # inshop 143, 3940, 3037
    (
        dbFeatImg,
        qFeatImg,
        dbFeatText,
        qFeatText,
        gt,
        numDb,
        numQ,
    ) = dataset.get_dataset_embeddings(embed_type=student_embedding)

    (
        dbFeatImg2,
        qFeatImg2,
        dbFeatText2,
        qFeatText2,
        gt2,
        numDb2,
        numQ2,
    ) = dataset.get_dataset_embeddings(embed_type=teacher_embedding)

    imgCCA = CCA(latent_dims=700)
    x = dbFeatImg - np.mean(dbFeatImg, axis=0)
    y = dbFeatImg2 - np.mean(dbFeatImg2, axis=0)
    imgCCA.fit((x, y))

    studentPca = PCA(n_components=700)
    _ = studentPca.fit_transform(dbFeatImg)

    TeacherPca = PCA(n_components=700)
    _ = TeacherPca.fit_transform(dbFeatImg2)

    pca_comp = 400  # The dimension size to evaluate

    cca_better = 0
    pca_better = 0
    gt_size = np.array([len(x) for x in gt])
    qidxlist = np.where(gt_size > 40)[0]
    bestidx = 0
    bestratio = 0

    for qidx in tqdm(qidxlist):

        qidx = 3940  # TODO: This is for plotting change later

        dbFeatImgPca = studentPca.transform(dbFeatImg)[:, :pca_comp]
        qFeatImgPca = studentPca.transform(qFeatImg)[:, :pca_comp]

        x = dbFeatImg - np.mean(dbFeatImg, axis=0)
        y = dbFeatImg2 - np.mean(dbFeatImg2, axis=0)
        dbFeatImgCCA, dbFeatTextCCA = imgCCA.transform(
            (x, y * 0)
        )  # multiplying by zero since we are only interested in the first part
        dbFeatImgCCA = dbFeatImgCCA[:, :pca_comp]

        x = qFeatImg - np.mean(qFeatImg, axis=0)
        y = qFeatImg2 - np.mean(qFeatImg2, axis=0)
        qFeatImgCCA, qFeatTextCCA = imgCCA.transform((x, y * 0))
        qFeatImgCCA = qFeatImgCCA[:, :pca_comp]

        # sklearn cosine similarity

        pcasim = cosine_similarity(
            qFeatImgPca[qidx].reshape(1, -1), dbFeatImgPca
        )
        pcapreds = np.argsort(pcasim, axis=1)[0][::-1][:n]

        ccasim = cosine_similarity(
            qFeatImgCCA[qidx].reshape(1, -1), dbFeatImgCCA
        )
        ccapreds = np.argsort(ccasim, axis=1)[0][::-1][:n]

        pca_correct = np.intersect1d(pcapreds, gt[qidx])
        cca_correct = np.intersect1d(ccapreds, gt[qidx])

        gtvals = gt[qidx]

        # Tsne plotting for PCA and CCA separately
        # selected_idxs = np.unique(np.hstack([pcapreds, ccapreds[:2]]))
        selected_idxs = pcapreds

        # selected_idxs label = 0 if its in gtvals
        labels = np.zeros(len(selected_idxs))
        for i, idx in enumerate(selected_idxs):
            if idx in gtvals:
                labels[i] = 1
        # labels[np.isin(selected_idxs, gtvals)] = 1
        # tsne for pca

        perplexity = 8
        pca_tsne = TSNE(
            n_components=2,
            random_state=0,
            perplexity=perplexity,
            metric="cosine",
        )
        pca_tsne_results = pca_tsne.fit_transform(dbFeatImgPca[selected_idxs])

        # tsne for cca
        cca_tsne = TSNE(
            n_components=2,
            random_state=0,
            perplexity=perplexity,
            metric="cosine",
        )
        cca_tsne_results = cca_tsne.fit_transform(dbFeatImgCCA[selected_idxs])

        # two  subplots, one for pca and other for CCA
        fig, axs = plt.subplots(1, 2, figsize=(10, 5))
        axs[0].scatter(
            pca_tsne_results[:, 0],
            pca_tsne_results[:, 1],
            c=labels,
            cmap="viridis",
        )
        axs[0].set_title(f"PCA TSNE for query {qidx}")
        axs[1].scatter(
            cca_tsne_results[:, 0],
            cca_tsne_results[:, 1],
            c=labels,
            cmap="viridis",
        )
        axs[1].set_title(f"CCA TSNE for query {qidx}")
        plt.savefig(f"data/plots/tsne_{qidx}.png")

        # # plot pca tsne
        # plt.scatter(pca_tsne_results[:, 0], pca_tsne_results[:, 1], c=labels, cmap='viridis')
        # plt.title(f"PCA TSNE for query {qidx}")
        # plt.savefig(f"data/plots/pca_tsne_{qidx}.png")
        # plt.clf()

        # # plot cca tsne
        # plt.scatter(cca_tsne_results[:, 0], cca_tsne_results[:, 1], c=labels, cmap='viridis')
        # plt.title(f"CCA TSNE for query {qidx}")
        # plt.savefig(f"data/plots/cca_tsne_{qidx}.png")

        breakpoint()

        if len(cca_correct) / (len(pca_correct) + 1) > bestratio:
            bestratio = len(cca_correct) / (len(pca_correct) + 1)
            bestidx = qidx

        if len(cca_correct) > len(pca_correct):
            cca_better += 1
        elif len(cca_correct) < len(pca_correct):
            pca_better += 1

        if len(cca_correct) > len(pca_correct):
            print(f"CCA is better for query {qidx}")
            print(f"GT: {gtvals}")
            print(f"PCA: {pcapreds}")
            print(f"CCA: {ccapreds}")
            print("\n")

    breakpoint()


def measure_compute():
    processor = ViTImageProcessor.from_pretrained("facebook/dino-vits8")
    model = ViTModel.from_pretrained("facebook/dino-vits8")
    model = model.cuda()

    model_name = "facebook/dino-vits8"

    flops, macs, params = calculate_flops_hf(model_name=model_name)
    print(
        "%s FLOPs:%s  MACs:%s  Params:%s \n" % (model_name, flops, macs, params)
    )
    breakpoint()


def pca_ablations(
    k=3, embedding="dino", dataset_name="inshop", metric_name="map"
):
    # This is a map of dataset classes
    dataset_mapping = get_dataset_mapping()

    # creating an instance of the required dataset class
    dataset = dataset_mapping[dataset_name]()

    metrics_mapping = get_metrics_mapping()

    metric_function = metrics_mapping[metric_name]

    (
        dbFeatImg,
        qFeatImg,
        dbFeatText,
        qFeatText,
        gt,
        numDb,
        numQ,
    ) = dataset.get_dataset_embeddings(embed_type=embedding)

    # Setting up some params
    final_dim_size = dbFeatImg.shape[1]
    comp_range = [2**i for i in range(2, int(np.log2(final_dim_size)) + 1)] + [final_dim_size]
    
    # For Tokyo, we have a number of similar images in the db
    num_similar_db = None
    q_in_db = False
    if dataset_name == "tokyo_val":
        q_in_db = True
        num_similar_db = 6

    logger.info("Getting Predictions")

    results = {
        f"pca": [],
        f"pca_norm": [],
        f"pca_norm_random": [],
        f"norm_random": [],
        "comp_range": comp_range,
    }

    pca = PCA(n_components=comp_range[-1])
    _ = pca.fit_transform(dbFeatImg)

    # Getting the normalized variants of the original embeddings
    dbFeatImgNorm = (dbFeatImg - np.mean(dbFeatImg, axis=0)) / np.std(
        dbFeatImg, axis=0
    )
    qFeatImgNorm = (qFeatImg - np.mean(dbFeatImg, axis=0)) / np.std(
        dbFeatImg, axis=0
    )

    for pca_comp in tqdm(comp_range):
        logger.info(f"Embedding Dimension: {pca_comp}")

        ##### Original Embedding with Norm
        random_indices = np.random.choice(
            np.arange(comp_range[-1]), pca_comp, replace=False
        )
        recall = metric_function(
            qFeatImgNorm.copy()[:, random_indices],
            dbFeatImgNorm.copy()[:, random_indices],
            gt,
            numQ,
            k=k,
            q_in_db=q_in_db,
            num_similar_db=num_similar_db,
        )

        logger.info(f"{metric_name}@{k} for {embedding} images with Norm: = {recall}")
        results[f"norm_random"].append(recall)

        ##### PCA
        dbFeatImgPca = pca.transform(dbFeatImg)[:, :pca_comp]
        qFeatImgPca = pca.transform(qFeatImg)[:, :pca_comp]

        recall = metric_function(
            qFeatImgPca.copy(),
            dbFeatImgPca.copy(),
            gt,
            numQ,
            k=k,
            q_in_db=q_in_db,
            num_similar_db=num_similar_db,
        )
        logger.info(f"{metric_name}@{k} for {embedding} images with PCA: = {recall}")

        results[f"pca"].append(recall)

        ##### PCA Norm Random
        dbFeatImgPca = pca.transform(dbFeatImg)
        qFeatImgPca = pca.transform(qFeatImg)

        dbMean = np.mean(dbFeatImgPca, axis=0)
        dbStd = np.std(dbFeatImgPca, axis=0)

        dbFeatImgPca = (dbFeatImgPca - dbMean) / dbStd
        qFeatImgPca = (qFeatImgPca - dbMean) / dbStd

        # select pca_comp random indixes from 0 to final_dim_size
        random_indices = np.random.choice(
            np.arange(comp_range[-1]), pca_comp, replace=False
        )

        dbFeatImgPca = dbFeatImgPca[:, random_indices]
        qFeatImgPca = qFeatImgPca[:, random_indices]

        recall = metric_function(
            qFeatImgPca.copy(),
            dbFeatImgPca.copy(),
            gt,
            numQ,
            k=k,
            q_in_db=q_in_db,
            num_similar_db=num_similar_db,
        )
        logger.info(
            f"{metric_name}@{k} for {embedding} images with PCA_Norm_Random: = {recall}"
        )

        results[f"pca_norm_random"].append(recall)

        ##### PCA Norm
        dbFeatImgPca = pca.transform(dbFeatImg)
        qFeatImgPca = pca.transform(qFeatImg)

        dbMean = np.mean(dbFeatImgPca, axis=0)
        dbStd = np.std(dbFeatImgPca, axis=0)

        dbFeatImgPca = (dbFeatImgPca - dbMean) / dbStd
        qFeatImgPca = (qFeatImgPca - dbMean) / dbStd

        dbFeatImgPca = dbFeatImgPca[:, :pca_comp]
        qFeatImgPca = qFeatImgPca[:, :pca_comp]
        recall = metric_function(
            qFeatImgPca.copy(),
            dbFeatImgPca.copy(),
            gt,
            numQ,
            k=k,
            q_in_db=q_in_db,
        )
        logger.info(
            f"{metric_name}@{k} for {embedding} images with PCA_Norm: = {recall}"
        )

        results[f"pca_norm"].append(recall)

    # save results pickle
    with open(
        f"data/pkl/{dataset_name}_pca_ablations_{metric_name}@{k}_{embedding}.pkl",
        "wb",
    ) as f:
        pickle.dump(results, f)


def entropy_ablations(n=4, embedding="dino", dataset_name="inshop"):
    metrics_mapping = get_metrics_mapping()

    metric_function = metrics_mapping['recall']

    get_recall = metric_function

    dataset_mapping = get_dataset_mapping()

    dataset = dataset_mapping[dataset_name]()

    (
        dbFeatImg,
        qFeatImg,
        dbFeatText,
        qFeatText,
        gt,
        numDb,
        numQ,
    ) = dataset.get_dataset_embeddings(embed_type=embedding)

    final_dim_size = dbFeatImg.shape[1]

    comp_range = list(range(10, final_dim_size + 1, 50))

    # Other hyperparameters
    q_in_db = False

    logger.info("Getting Predictions")

    pca = PCA(n_components=comp_range[-1])
    _ = pca.fit_transform(dbFeatImg)

    pca_comp = comp_range[-1]
    print("Embedding Dimension:", pca_comp)

    hist_bins = 100
    topk = 6000

    ##### PCA Embeddings
    dbFeatImgPca = pca.transform(dbFeatImg)[:, :pca_comp]
    qFeatImgPca = pca.transform(qFeatImg)[:, :pca_comp]

    if os.path.exists(f"data/entropy_pca_{dataset_name}_{embedding}.npy"):
        sim_space = np.load(f"data/entropy_pca_{dataset_name}_{embedding}.npy")
    else:
        sim_space = torch.zeros((dbFeatImgPca.shape[0], dbFeatImgPca.shape[0]))
        for i in tqdm(range(dbFeatImgPca.shape[0])):
            sim_space[i] = F.cosine_similarity(
                torch.tensor(dbFeatImgPca[i]).unsqueeze(0),
                torch.tensor(dbFeatImgPca),
                dim=1,
            )
        sim_space = sim_space.numpy()
        # save sim space
        np.save(f"data/entropy_pca_{dataset_name}_{embedding}.npy", sim_space)

    hist = [
        np.histogram(
            sim_space[i], range=(-1, 1), density=False, bins=hist_bins
        )[0]
        for i in range(sim_space.shape[0])
    ]
    sim_space_stack = sim_space.flatten()
    sim_space_topk = np.sort(sim_space, axis=1)[:, -topk:]
    sim_space_topk = sim_space_topk.flatten()
    entropy = [scipy.stats.entropy(hist[i]) for i in range(sim_space.shape[0])]

    print("Mean Entropy:", np.mean(entropy))

    recall = get_recall(
        qFeatImgPca.copy(),
        dbFeatImgPca.copy(),
        gt,
        numQ,
        k=n,
        q_in_db=q_in_db,
    )
    print(f"Recall@{n} for {embedding} images with PCA: = {recall}")

    # PCA Norm
    dbFeatImgPca = pca.transform(dbFeatImg)
    qFeatImgPca = pca.transform(qFeatImg)

    dbMean = np.mean(dbFeatImgPca, axis=0)
    dbStd = np.std(dbFeatImgPca, axis=0)

    dbFeatImgPca = (dbFeatImgPca - dbMean) / dbStd
    qFeatImgPca = (qFeatImgPca - dbMean) / dbStd

    dbFeatImgPca = dbFeatImgPca[:, :pca_comp]
    qFeatImgPca = qFeatImgPca[:, :pca_comp]

    if os.path.exists(f"data/entropy_pca_norm_{dataset_name}_{embedding}.npy"):
        sim_space = np.load(f"data/entropy_pca_norm_{dataset_name}_{embedding}.npy")
    else:
        sim_space = torch.zeros((dbFeatImgPca.shape[0], dbFeatImgPca.shape[0]))
        for i in tqdm(range(dbFeatImgPca.shape[0])):
            sim_space[i] = F.cosine_similarity(
                torch.tensor(dbFeatImgPca[i]).unsqueeze(0),
                torch.tensor(dbFeatImgPca),
                dim=1,
            )
        sim_space = sim_space.numpy()
        np.save(f"data/entropy_pca_norm_{dataset_name}_{embedding}.npy", sim_space)

    hist = [
        np.histogram(
            sim_space[i], range=(-1, 1), density=False, bins=hist_bins
        )[0]
        for i in range(sim_space.shape[0])
    ]
    sim_space_stack_norm = sim_space.flatten()
    sim_space_norm_topk = np.sort(sim_space, axis=1)[:, -topk:]
    sim_space_norm_topk = sim_space_norm_topk.flatten()
    entropy = [scipy.stats.entropy(hist[i]) for i in range(sim_space.shape[0])]

    print("Mean Entropy:", np.mean(entropy))

    set_plot_properties()

    fig, ax = plt.subplots(1, 1, figsize=(6, 4))


    embedding_map = {
        "clip": "CLIP",
        "vit": "ViT",
        "dinob16": "Dino",
        "dinov2_large": "DINO-V2-Large",
        "dinos8": "DINO-S8",
        "cosplace": "Cosplace",
    }

    print('Plotting...')

    plot_several_pdf(
        data_vector_list=[sim_space_topk, sim_space_norm_topk],
        legend=["PCA", "VA-PCA"],
        title_str=f"{embedding_map[embedding]}",
        xlabel="Similarity",
        ylabel="Frequency",
        ax=ax,
    )

    plt.savefig(f"data/plots/entropy_pca_vs_pca_norm_{dataset_name}_{embedding}.png", bbox_inches='tight')


def kl_divergence(pmf1, pmf2):
    # calculate only for the non-zero values

    kl_div = 0
    for i in range(len(pmf1)):
        if pmf1[i] == 0 or pmf2[i] == 0:
            continue
        kl_div += pmf1[i] * np.log(pmf1[i] / pmf2[i])

    return kl_div


def entropy_dimensions_ablations(n=3, embedding="dino", dataset="sop"):
    # This is a map of dataset classes
    dataset_mapping = get_dataset_mapping()

    # creating an instance of the required dataset class
    dataset = dataset_mapping[dataset]()

    (
        dbFeatImg,
        qFeatImg,
        dbFeatText,
        qFeatText,
        gt,
        numDb,
        numQ,
    ) = dataset.get_dataset_embeddings(embed_type=embedding)

    final_dim_size = dbFeatImg.shape[1]

    comp_range = list(range(10, final_dim_size + 1, 50))

    # Other hyperparameters
    q_in_db = False

    logger.info("Getting Predictions")

    pca = PCA(n_components=comp_range[-1])
    _ = pca.fit_transform(dbFeatImg)

    pca_comp = comp_range[-1]
    print("Embedding Dimension:", pca_comp)

    hist_bins = 20
    topk = 1
    # numq = qFeatImg.shape[0]
    numq_entropy = 1000
    num_dimensions = -1

    pmf_list = []
    pmf_norm_list = []
    for num_dimensions in [50, 200]:
        ##### PCA Embeddings
        dbFeatImgPca = pca.transform(dbFeatImg)[:, :pca_comp]
        qFeatImgPca = pca.transform(qFeatImg)[:, :pca_comp]

        sim_space = torch.zeros((numq_entropy, dbFeatImgPca.shape[0]))
        for i in tqdm(range(numq_entropy)):
            sim_space[i] = F.cosine_similarity(
                torch.tensor(qFeatImgPca[i][:num_dimensions]).unsqueeze(0),
                torch.tensor(dbFeatImgPca[:, :num_dimensions]),
                dim=1,
            )
        sim_space = sim_space.numpy()

        hist = [
            np.histogram(
                sim_space[i], range=(-1, 1), density=False, bins=hist_bins
            )[0]
            for i in range(sim_space.shape[0])
        ]
        sim_space_stack = sim_space.flatten()
        sim_space_topk = np.sort(sim_space, axis=1)[:, -topk:]
        sim_space_topk = sim_space_topk.flatten()
        topk_hist = np.histogram(
            sim_space_topk, range=(-1, 1), density=False, bins=hist_bins
        )[0]
        topk_pmf = topk_hist / np.sum(topk_hist)
        pmf_list.append(topk_pmf.copy())
        entropy = [
            scipy.stats.entropy(hist[i]) for i in range(sim_space.shape[0])
        ]

        print("Mean Entropy:", np.mean(entropy))

        recall = get_recall(
            qFeatImgPca.copy(),
            dbFeatImgPca.copy(),
            gt,
            numQ,
            n=n,
            q_in_db=q_in_db,
        )
        print(f"Recall@{n} for {embedding} images with PCA: = {recall}")

        # PCA Norm
        dbFeatImgPca = pca.transform(dbFeatImg)
        qFeatImgPca = pca.transform(qFeatImg)

        dbMean = np.mean(dbFeatImgPca, axis=0)
        dbStd = np.std(dbFeatImgPca, axis=0)

        dbFeatImgPca = (dbFeatImgPca - dbMean) / dbStd
        qFeatImgPca = (qFeatImgPca - dbMean) / dbStd

        dbFeatImgPca = dbFeatImgPca[:, :pca_comp]
        qFeatImgPca = qFeatImgPca[:, :pca_comp]

        sim_space = torch.zeros((numq_entropy, dbFeatImgPca.shape[0]))
        for i in tqdm(range(numq_entropy)):
            sim_space[i] = F.cosine_similarity(
                torch.tensor(qFeatImgPca[i][:num_dimensions]).unsqueeze(0),
                torch.tensor(dbFeatImgPca[:, :num_dimensions]),
                dim=1,
            )
        sim_space = sim_space.numpy()

        hist = [
            np.histogram(
                sim_space[i], range=(-1, 1), density=False, bins=hist_bins
            )[0]
            for i in range(sim_space.shape[0])
        ]
        sim_space_stack_norm = sim_space.flatten()
        sim_space_norm_topk = np.sort(sim_space, axis=1)[:, -topk:]
        sim_space_norm_topk = sim_space_norm_topk.flatten()
        topk_norm_hist = np.histogram(
            sim_space_norm_topk, range=(-1, 1), density=False, bins=hist_bins
        )[0]
        topk_norm_pmf = topk_norm_hist / np.sum(topk_norm_hist)
        pmf_norm_list.append(topk_norm_pmf.copy())
        entropy = [
            scipy.stats.entropy(hist[i]) for i in range(sim_space.shape[0])
        ]

        print("Mean Entropy:", np.mean(entropy))

    pmf_list = np.array(pmf_list)
    pmf_norm_list = np.array(pmf_norm_list)

    kl_div = kl_divergence(pmf_list[0], pmf_list[1])
    kl_div_norm = kl_divergence(pmf_norm_list[0], pmf_norm_list[1])

    print(f"KL Divergence: {kl_div}")
    print(f"KL Divergence Norm: {kl_div_norm}")

    # fig, ax = plt.subplots(1, 1, figsize=(10, 7))

    # sim_space_stack_norm = sim_space.flatten()
    # plot_several_pdf(data_vector_list=[sim_space_stack, sim_space_stack_norm],
    #     legend=['PCA', 'PCA Norm'],
    #     title_str='Cosine Similarity Distribution',
    #     xlabel='Similarity',
    #     ylabel='Frequency',
    #     ax=ax)

    # plt.savefig(f"data/plots/entropy_pca_vs_pca_norm_stack_dim_{num_dimensions}.png")


def main():
    all_embeddings = [
        "dinov2_vitg14",
        "dinob8",
        "dinos8",
        "dinov2_small",
        "dinov2_large",
        "vit",
        "clip",
    ]

    dataset_names = ["inshop", "tokyo_val", "sop", "pitts"]
    k = 1
    for dataset_name in dataset_names:
        for embedding in all_embeddings:
            pca_ablations(k=k, embedding=embedding, dataset_name=dataset_name, metric_name="recall")
            plot_pca_ablations(
                k=k, embedding=embedding, dataset_name=dataset_name, metric_name="recall"
            )


if __name__ == "__main__":
    cfg = singletons.get_cfg()
    print(
        "\n######\nRunning With The Following Conf:\n", OmegaConf.to_yaml(cfg)
    )
    logger = singletons.get_logger()
    # cca_tsne(n=10, student_embedding="dino", teacher_embedding="dinov2")
    # cca_zoo(n=4, student_embedding="dinov2_small", teacher_embedding="dinob8")
    # plot_CCA(k=4, student_embedding="dinov2_small", teacher_embedding="dinob8")

    # dcca(n=4)

    # get_embeddings.clip_imgs(None, None)

    # measure_compute()

    # main()

    entropy_ablations(n=4, embedding="cosplace", dataset_name="pitts") 
    # entropy_dimensions_ablations(n=4, embedding="dinob8")

"""
0.213
0.223

0.139
"""
