from abc import ABC, abstractmethod
from typing import Any
from tirr.utils import singletons
from tirr.utils.get_embeddings import get_embedding_function_mapping
from tirr.utils import get_embeddings
from tirr.packages.netvlad import pittsburgh, tokyo247
import numpy as np
import pickle
import os
import torchvision
from torch.utils.data import Dataset
from tirr.utils.encoders import CLIPImg, CLIPText
from tirr.MetricDatasets.stanford_products import StanfordOnlineProducts
from tirr.MetricDatasets.CUB import Cub200
from tirr.MetricDatasets.inshop import InShop
from tqdm import tqdm
import pandas
from cca_zoo.models import CCA
from sklearn.metrics.pairwise import cosine_similarity
import torch
from sklearn.decomposition import PCA
from torch.nn import functional as F


cfg = singletons.get_cfg()
logger = singletons.get_logger()


class PRDataset(ABC, Dataset):
    required_variables = ["dir", "text_desc_file"]

    @abstractmethod
    def __init__(self):
        for var in self.required_variables:
            if not hasattr(self, var):
                raise NotImplementedError(
                    f"Class {self.__class__.__name__} must implement {var}"
                )

    @abstractmethod
    def get_dataset_embeddings(self):
        pass


class Dataset17P(PRDataset):
    def __init__(self):
        self.dir = "/store/ /datasets/17places/"
        self.text_desc_file = self.dir + "text_descriptions_17P.pkl"

        gt_17P = np.load(
            f"{self.dir}my_ground_truth_new.npy", allow_pickle=True
        )
        self.gt = []
        for x in gt_17P:
            self.gt.append(np.array(x[1]))

        self.numQ = 406
        self.numDb = 406

        self.embedding_function_mapping = get_embedding_function_mapping()

    def get_dataset_embeddings(self, embed_type="clip"):
        logger.info("Getting Dataset Embeddings")

        image_emb_file = self.dir + f"image_embeddings_17P_{embed_type}.npy"
        emb_function = self.embedding_function_mapping[embed_type]

        ref_img_files = [
            f'{self.dir + "ref/" + str(i)}.jpg' for i in range(0, 406)
        ]  # ref images
        query_img_files = [
            f'{self.dir + "query/" + str(i)}.jpg' for i in range(0, 406)
        ]  # query images
        all_img_files = ref_img_files + query_img_files

        # load text descriptions for equal spread dataset
        with open(self.text_desc_file, "rb") as f:
            text_descriptions = pickle.load(f)
        text_descriptions = [x[1] for x in text_descriptions]

        # Get embeddings will return normalized embeddings
        if not os.path.exists(image_emb_file):
            img_emb = emb_function(all_img_files)
            np.save(image_emb_file, img_emb)
        else:
            img_emb = np.load(image_emb_file)
        text_emg = get_embeddings.clip_text(text_descriptions, batch_size=72)
        text_emg = get_embeddings.clip_text(text_descriptions)

        dbFiles = all_img_files[:406]
        qFiles = all_img_files[406:]
        indices = list(np.arange(len(all_img_files)))
        dbIndices = indices[:406]
        qIndices = indices[406:]

        self.dbFeatImg = img_emb[dbIndices].astype("float32")
        self.qFeatImg = img_emb[qIndices].astype("float32")

        self.dbFeatText = text_emg[dbIndices].astype("float32")
        self.qFeatText = text_emg[qIndices].astype("float32")

        return (
            self.dbFeatImg,
            self.qFeatImg,
            self.dbFeatText,
            self.qFeatText,
            self.gt,
            self.numDb,
            self.numQ,
        )


class DatasetPitts(PRDataset):
    def __init__(self, no_training=True, embed_type="dino"):
        self.dir = "/store/ /datasets/pitts250k/"
        self.text_desc_file = self.dir + "text_descriptions_pitts30k.pkl"

        dataset = pittsburgh.get_whole_val_set()
        self.gt = dataset.getPositives()
        self.numQ = dataset.dbStruct.numQ
        self.numDb = dataset.dbStruct.numDb
        self.embedding_function_mapping = get_embedding_function_mapping()
        if not no_training:
            self.emb_type = embed_type
            # adding suffix 1 here because get_dataset_embeddings also creates a self.dbFeatImg
            (
                self.dbFeatImg1,
                self.qFeatImg1,
                self.dbFeatText1,
                self.qFeatText1,
                _,
                _,
                _,
            ) = self.get_dataset_embeddings(embed_type=embed_type)

            self.final_dim_size = self.dbFeatImg1.shape[1]

            pca = PCA(n_components=self.final_dim_size)
            pca.fit(self.dbFeatImg1)
            self.dbFeatImg1PCA = pca.transform(self.dbFeatImg1)
            self.qFeatImg1PCA = pca.transform(self.qFeatImg1)

            dbMean = np.mean(self.dbFeatImg1PCA, axis=0)
            dbStd = np.std(self.dbFeatImg1PCA, axis=0)

            self.dbFeatImg1PCA = (self.dbFeatImg1PCA - dbMean) / dbStd
            self.qFeatImg1PCA = (self.qFeatImg1PCA - dbMean) / dbStd


            # calculate this only when sim_space is not present
            if not os.path.exists('data/pkl/' + f"{self.__class__.__name__}_teacher_sim_space_{embed_type}.npy"):
                # Calculate cosine similarity one row at a time
                print('Calculating Teacher Sim Space')
                self.teacher_sim_space = torch.zeros((self.dbFeatImg1PCA.shape[0], self.dbFeatImg1PCA.shape[0]))
                for i in tqdm(range(self.teacher_sim_space.shape[0])):
                    self.teacher_sim_space[i] = F.cosine_similarity(torch.tensor(self.dbFeatImg1PCA[i]).unsqueeze(0), torch.tensor(self.dbFeatImg1PCA), dim=1)

                # save teacher_sim_space
                np.save('data/pkl/' + f"{self.__class__.__name__}_teacher_sim_space_{embed_type}.npy", self.teacher_sim_space.numpy())
            else:
                self.teacher_sim_space = torch.tensor(np.load('data/pkl/' + f"{self.__class__.__name__}_teacher_sim_space_{embed_type}.npy"))

            self.teacher_topk_indices = self.teacher_sim_space.topk(dim=1, k=500)[1]
            self.teacher_topk_values = self.teacher_sim_space.topk(dim=1, k=500)[0]


        

    def get_dataset_embeddings(self, embed_type="clip"):
        image_emb_file = (
            self.dir + f"image_embeddings_pitts30k_{embed_type}.npy"
        )
        emb_function = self.embedding_function_mapping[embed_type]

        with open(self.text_desc_file, "rb") as f:
            pittsdata = pickle.load(f)

        all_img_files = np.array([x[0] for x in pittsdata])
        text_descriptions = [x[1] for x in pittsdata]

        # replace 'TextMapReduce' in all_img_files with 'datasets'
        all_img_files = np.array(
            [x.replace("TextMapReduce", "datasets") for x in all_img_files]
        )

        # Get embeddings will return normalized embeddings
        if not os.path.exists(image_emb_file):
            img_emb = emb_function(all_img_files)
            np.save(image_emb_file, img_emb)
        else:
            img_emb = np.load(image_emb_file)

        # text_emg = get_embeddings.clip_text(text_descriptions, batch_size=72)

        dbFiles = all_img_files[: self.numDb]
        qFiles = all_img_files[self.numDb :]
        indices = list(np.arange(len(all_img_files)))
        dbIndices = indices[: self.numDb]
        qIndices = indices[self.numDb :]

        self.dbFeatImg = img_emb[dbIndices].astype("float32")
        self.qFeatImg = img_emb[qIndices].astype("float32")
        # self.dbFeatText = text_emg[dbIndices].astype("float32")
        # self.qFeatText = text_emg[qIndices].astype("float32")

        self.dbFeatText = None
        self.qFeatText = None

        return (
            self.dbFeatImg,
            self.qFeatImg,
            self.dbFeatText,
            self.qFeatText,
            self.gt,
            self.numDb,
            self.numQ,
        )

    def __getitem__(self, index) -> Any:
        indices = self.teacher_topk_indices[index]
        img_emb1 = torch.tensor(self.dbFeatImg1PCA[indices])
        values = self.teacher_topk_values[index]

        return img_emb1, values

    
    def __len__(self) -> int:
        return self.numDb

        

class DatasetPittsTrain(PRDataset):
    def __init__(self, no_training=True, pos_count=2, neg_count=10):
        self.dir = "/store/ /datasets/pitts250k/"
        self.text_desc_file = self.dir + "text_descriptions_pitts30k_train.pkl"
        dataset = pittsburgh.get_whole_training_set()
        self.all_img_files, self.gt, self.numQ, self.numDb = self.get_metadata()
        self.embedding_function_mapping = get_embedding_function_mapping()

        with open(self.text_desc_file, "rb") as f:
            textdata = pickle.load(f)

        self.text_descriptions = [x[1] for x in textdata]

        # For Triplet Loss
        self.pos_count = pos_count
        self.neg_count = neg_count

        if not no_training:
            (
                self.dbFeatImg,
                self.qFeatImg,
                self.dbFeatText,
                self.qFeatText,
                _,
                _,
                _,
            ) = self.get_dataset_embeddings(embed_type="clip")

    def __getitem__(self, index) -> Any:
        query_img_file = self.qImgFiles[index]

        # select self.pos_count positive idxs
        pos_idx = np.random.choice(
            self.gt[index], self.pos_count, replace=False
        )
        pos_img_files = [self.dbImgFiles[i] for i in pos_idx]

        all_db_idxs = np.arange(self.numDb)
        neg_idx = np.setdiff1d(all_db_idxs, self.gt[index])
        neg_idx = np.random.choice(neg_idx, self.neg_count, replace=False)
        neg_img_files = [self.dbImgFiles[i] for i in neg_idx]

        query_text = self.text_descriptions[index + self.numDb]
        pos_text = [self.text_descriptions[i] for i in pos_idx]
        neg_text = [self.text_descriptions[i] for i in neg_idx]

        query_img_emb = np.expand_dims(self.qFeatImg[index], axis=0)
        pos_img_emb = self.dbFeatImg[pos_idx]
        neg_img_emb = self.dbFeatImg[neg_idx]

        query_text_emb = np.expand_dims(self.qFeatText[index], axis=0)
        pos_text_emb = self.dbFeatText[pos_idx]
        neg_text_emb = self.dbFeatText[neg_idx]

        return (
            query_img_emb,
            pos_img_emb,
            neg_img_emb,
            query_text_emb,
            pos_text_emb,
            neg_text_emb,
        )

    def __len__(self) -> int:
        return self.numQ

    def get_metadata(self):
        dataset = pittsburgh.get_whole_training_set()
        gt = dataset.getPositives()
        numQ = dataset.dbStruct.numQ
        numDb = dataset.dbStruct.numDb
        qImgFiles = dataset.dbStruct.qImage
        dbImgFiles = dataset.dbStruct.dbImage

        self.qImgFiles = [self.dir + x for x in qImgFiles]
        self.dbImgFiles = [self.dir + x for x in dbImgFiles]

        all_img_files = self.dbImgFiles + self.qImgFiles

        return all_img_files, gt, numQ, numDb

    def get_dataset_embeddings(self, embed_type="clip"):
        image_emb_file = (
            self.dir + f"image_embeddings_pitts30k_{embed_type}_train.npy"
        )
        emb_function = self.embedding_function_mapping[embed_type]

        with open(self.text_desc_file, "rb") as f:
            pittsdata = pickle.load(f)

        all_img_files = np.array([x[0] for x in pittsdata])
        text_descriptions = [x[1] for x in pittsdata]

        # replace 'TextMapReduce' in all_img_files with 'datasets'
        all_img_files = np.array(
            [x.replace("TextMapReduce", "datasets") for x in all_img_files]
        )

        # Get embeddings will return normalized embeddings
        if not os.path.exists(image_emb_file):
            img_emb = emb_function(all_img_files)
            np.save(image_emb_file, img_emb)
        else:
            img_emb = np.load(image_emb_file)
        # text_emg = get_embeddings.clip_text(text_descriptions, batch_size=72)
        dbFiles = all_img_files[: self.numDb]
        qFiles = all_img_files[self.numDb :]
        indices = list(np.arange(len(all_img_files)))
        dbIndices = indices[: self.numDb]
        qIndices = indices[self.numDb :]

        self.dbFeatImg = img_emb[dbIndices].astype("float32")
        self.qFeatImg = img_emb[qIndices].astype("float32")
        # self.dbFeatText = text_emg[dbIndices].astype("float32")
        # self.qFeatText = text_emg[qIndices].astype("float32")

        self.dbFeatText = None
        self.qFeatText = None

        return (
            self.dbFeatImg,
            self.qFeatImg,
            self.dbFeatText,
            self.qFeatText,
            self.gt,
            self.numDb,
            self.numQ,
        )


class DatasetKITTI(PRDataset):
    def __init__(self):
        self.dir = "/store/ /datasets/KITTI/"
        self.text_desc_file = self.dir + "text_descriptions_kitti.pkl"

        _, self.gt, self.numQ, self.numDb = self.get_metadata()
        self.embedding_function_mapping = get_embedding_function_mapping()

    def get_metadata(self):
        sequence_list = ["00", "02", "05", "06", "07", "08"]
        gt = []
        all_db_img_files = []
        all_q_img_files = []
        numDB = 0
        numQ = 0

        for sequence in sequence_list:
            (
                img_files,
                seq_gt,
                seq_numQ,
                seq_numDB,
            ) = self.get_metadata_one_sequence(sequence)
            all_db_img_files += img_files[:seq_numDB]
            all_q_img_files += img_files[seq_numDB:]

            for i in range(len(seq_gt)):
                for j in range(len(seq_gt[i])):
                    seq_gt[i][j] += numDB

            gt += seq_gt

            numDB += seq_numDB
            numQ += seq_numQ

        all_img_files = all_db_img_files + all_q_img_files

        return all_img_files, gt, numQ, numDB

    def get_metadata_one_sequence(self, sequence="00"):
        pair_file = self.dir + "pairs/pairs_kitti/neg_1/" + sequence + ".txt"

        with open(pair_file, "r") as f:
            pairs = f.readlines()
            pairs = [pair.strip().split() for pair in pairs]

        pairs = np.array(pairs)
        pairs = pairs[pairs[:, 2] == "1"][:, :2]

        gt_dict = {}
        for pair in pairs:
            if pair[0] not in gt_dict:
                gt_dict[pair[0]] = []
            gt_dict[pair[0]].append(pair[1])

        unq_imgs = {}
        unq_idx_db = 0
        all_img_files = []
        gt = []

        for imgs in gt_dict.values():
            for img in imgs:
                if img not in unq_imgs:
                    unq_imgs[img] = unq_idx_db
                    unq_idx_db += 1
                    all_img_files.append(
                        self.dir
                        + "dataset/sequences/"
                        + sequence
                        + "/image_2/"
                        + img
                        + ".png"
                    )

        numDB = len(all_img_files)
        for img in gt_dict.keys():
            if img not in unq_imgs:
                unq_imgs[img] = []
                all_img_files.append(
                    self.dir
                    + "dataset/sequences/"
                    + sequence
                    + "/image_2/"
                    + img
                    + ".png"
                )
                db_imgs = gt_dict[img]
                gt.append([unq_imgs[db_img] for db_img in db_imgs])

        numQ = len(all_img_files) - numDB

        return all_img_files, gt, numQ, numDB

    def get_extra_kitti_db_images(self):
        dir = "/store/ /datasets/KITTI/dataset/sequences/19/image_2/"
        files = os.listdir(dir)
        caption = "Big fat panda"
        files = [[dir + file, caption] for file in files]
        return files, len(files)

    def get_dataset_embeddings(self, embed_type="clip"):
        image_emb_file = (
            self.dir + f"image_embeddings_pitts30k_{embed_type}_train.npy"
        )
        emb_function = self.embedding_function_mapping[embed_type]

        with open(self.text_desc_file, "rb") as f:
            textdata = pickle.load(f)

        # # Add pitsburgh data in map
        # pittsaddon = 50_000
        # with open("/store/ /datasets/pitts250k/text_descriptions_pitts30k.pkl", "rb") as f:
        #     pittsdata = pickle.load(f)

        kitii_extra_db, addon = self.get_extra_kitti_db_images()

        textdata = (
            textdata[: self.numDb] + kitii_extra_db + textdata[self.numDb :]
        )

        self.numDb += addon

        all_img_files = np.array([x[0] for x in textdata])
        text_descriptions = [x[1] for x in textdata]

        # replace 'TextMapReduce' in all_img_files with 'datasets'
        all_img_files = np.array(
            [x.replace("TextMapReduce", "datasets") for x in all_img_files]
        )

        # Get embeddings will return normalized embeddings
        if not os.path.exists(image_emb_file):
            img_emb = emb_function(all_img_files)
            np.save(image_emb_file, img_emb)
        else:
            img_emb = np.load(image_emb_file)
        text_emg = get_embeddings.clip_text(text_descriptions, batch_size=72)

        dbFiles = all_img_files[: self.numDb]
        qFiles = all_img_files[self.numDb :]
        indices = list(np.arange(len(all_img_files)))
        dbIndices = indices[: self.numDb]
        qIndices = indices[self.numDb :]

        self.dbFeatImg = img_emb[dbIndices].astype("float32")
        self.qFeatImg = img_emb[qIndices].astype("float32")
        self.dbFeatText = text_emg[dbIndices].astype("float32")
        self.qFeatText = text_emg[qIndices].astype("float32")

        return (
            self.dbFeatImg,
            self.qFeatImg,
            self.dbFeatText,
            self.qFeatText,
            self.gt,
            self.numDb,
            self.numQ,
        )


class DatasetGoogleLMRetrival(PRDataset):
    def __init__(self):
        self.dir = "/store/ /datasets/GoogleLandmarks/retrival/"
        self.text_desc_file = (
            self.dir + "text_descriptions_googlelm_retrival.pkl"
        )

        _, self.gt, self.numQ, self.numDb = self.get_metadata()
        self.embedding_function_mapping = get_embedding_function_mapping()

    def get_metadata(self):
        with open(self.dir + "data.pkl", "rb") as f:
            data = pickle.load(f)

        all_img_files = data["img_files"]
        gt = data["gt"]
        numQ = data["numQ"]
        numDB = data["numDb"]

        all_img_files = [
            "/store/ /datasets/GoogleLandmarks/" + x for x in all_img_files
        ]

        return all_img_files, gt, numQ, numDB

    def get_dataset_embeddings(self, embed_type="clip"):
        image_emb_file = (
            self.dir
            + f"image_embeddings_googlelm_retrival_{embed_type}_train.npy"
        )
        emb_function = self.embedding_function_mapping[embed_type]

        with open(self.text_desc_file, "rb") as f:
            pittsdata = pickle.load(f)

        all_img_files = np.array([x[0] for x in pittsdata])
        text_descriptions = [x[1] for x in pittsdata]
        breakpoint()

        # Get embeddings will return normalized embeddings
        if not os.path.exists(image_emb_file):
            img_emb = emb_function(all_img_files)
            np.save(image_emb_file, img_emb)
        else:
            img_emb = np.load(image_emb_file)
        text_emg = get_embeddings.clip_text(text_descriptions, batch_size=72)
        dbFiles = all_img_files[: self.numDb]
        qFiles = all_img_files[self.numDb :]
        indices = list(np.arange(len(all_img_files)))
        dbIndices = indices[: self.numDb]
        qIndices = indices[self.numDb :]

        self.dbFeatImg = img_emb[dbIndices].astype("float32")
        self.qFeatImg = img_emb[qIndices].astype("float32")
        self.dbFeatText = text_emg[dbIndices].astype("float32")
        self.qFeatText = text_emg[qIndices].astype("float32")

        return (
            self.dbFeatImg,
            self.qFeatImg,
            self.dbFeatText,
            self.qFeatText,
            self.gt,
            self.numDb,
            self.numQ,
        )


class DatasetTokyoVal(PRDataset):
    def __init__(self, no_training=True, pos_count=2, neg_count=10):
        self.dir = "/store/ /datasets/tokyoTM/"
        self.text_desc_file = self.dir + "text_descriptions_tokyoTM_val.pkl"
        self.all_img_files, self.gt, self.numQ, self.numDb = self.get_metadata()
        self.embedding_function_mapping = get_embedding_function_mapping()

        # with open(self.text_desc_file, "rb") as f:
        #     textdata = pickle.load(f)

        # self.text_descriptions = [x[1] for x in textdata]

        # For Triplet Loss
        self.pos_count = pos_count
        self.neg_count = neg_count

        if not no_training:
            (
                self.dbFeatImg,
                self.qFeatImg,
                self.dbFeatText,
                self.qFeatText,
                _,
                _,
                _,
            ) = self.get_dataset_embeddings(embed_type="clip")

    def __getitem__(self, index) -> Any:
        query_img_file = self.qImgFiles[index]

        # select self.pos_count positive idxs
        pos_idx = np.random.choice(
            self.gt[index], self.pos_count, replace=False
        )
        pos_img_files = [self.dbImgFiles[i] for i in pos_idx]

        all_db_idxs = np.arange(self.numDb)
        neg_idx = np.setdiff1d(all_db_idxs, self.gt[index])
        neg_idx = np.random.choice(neg_idx, self.neg_count, replace=False)
        neg_img_files = [self.dbImgFiles[i] for i in neg_idx]

        query_text = self.text_descriptions[index + self.numDb]
        pos_text = [self.text_descriptions[i] for i in pos_idx]
        neg_text = [self.text_descriptions[i] for i in neg_idx]

        query_img_emb = np.expand_dims(self.qFeatImg[index], axis=0)
        pos_img_emb = self.dbFeatImg[pos_idx]
        neg_img_emb = self.dbFeatImg[neg_idx]

        query_text_emb = np.expand_dims(self.qFeatText[index], axis=0)
        pos_text_emb = self.dbFeatText[pos_idx]
        neg_text_emb = self.dbFeatText[neg_idx]

        return (
            query_img_emb,
            pos_img_emb,
            neg_img_emb,
            query_text_emb,
            pos_text_emb,
            neg_text_emb,
        )

    def __len__(self) -> int:
        return self.numQ

    def get_metadata(self):
        dataset = tokyo247.get_whole_val_set()
        gt = dataset.getPositives()
        numQ = dataset.dbStruct.numQ
        numDb = dataset.dbStruct.numDb
        qImgFiles = dataset.dbStruct.qImage
        dbImgFiles = dataset.dbStruct.dbImage

        self.qImgFiles = [self.dir + 'images/' + x for x in qImgFiles]
        self.dbImgFiles = [self.dir + 'images/' + x for x in dbImgFiles]

        all_img_files = self.dbImgFiles + self.qImgFiles

        return all_img_files, gt, numQ, numDb

    def get_dataset_embeddings(self, embed_type="clip"):
        image_emb_file = (
            self.dir + f"image_embeddings_tokyoTM_{embed_type}_val.npy"
        )
        emb_function = self.embedding_function_mapping[embed_type]

        # with open(self.text_desc_file, "rb") as f:
        #     pittsdata = pickle.load(f)

        # all_img_files = np.array([x[0] for x in pittsdata])
        # text_descriptions = [x[1] for x in pittsdata]

        # # replace 'TextMapReduce' in all_img_files with 'datasets'
        # all_img_files = np.array(
        #     [x.replace("TextMapReduce", "datasets") for x in all_img_files]
        # )

        all_img_files = self.all_img_files

        # Get embeddings will return normalized embeddings
        if not os.path.exists(image_emb_file):
            img_emb = emb_function(all_img_files)
            np.save(image_emb_file, img_emb)
        else:
            img_emb = np.load(image_emb_file)
        # text_emg = get_embeddings.clip_text(text_descriptions, batch_size=72)
        dbFiles = all_img_files[: self.numDb]
        qFiles = all_img_files[self.numDb :]
        indices = list(np.arange(len(all_img_files)))
        dbIndices = indices[: self.numDb]
        qIndices = indices[self.numDb :]

        self.dbFeatImg = img_emb[dbIndices].astype("float32")
        self.qFeatImg = img_emb[qIndices].astype("float32")
        # self.dbFeatText = text_emg[dbIndices].astype("float32")
        # self.qFeatText = text_emg[qIndices].astype("float32")

        self.dbFeatText = None
        self.qFeatText = None

        return (
            self.dbFeatImg,
            self.qFeatImg,
            self.dbFeatText,
            self.qFeatText,
            self.gt,
            self.numDb,
            self.numQ,
        )


class DatasetSOP(PRDataset):
    def __init__(self, split="test", no_training=True, embed_type="dino"):
        self.dir = "/store/ /datasets/Stanford_Online_Products/"
        self.text_desc_file = self.dir + "text_descriptions_SOP.pkl"
        self.metric_dataset = StanfordOnlineProducts(train = True if split == "train" else False)


        _, self.gt, self.numQ, self.numDb = self.get_metadata()
        self.embedding_function_mapping = get_embedding_function_mapping()
        self.split = split

        if not no_training:
            self.emb_type = embed_type
            # adding suffix 1 here because get_dataset_embeddings also creates a self.dbFeatImg
            (
                self.dbFeatImg1,
                self.qFeatImg1,
                self.dbFeatText1,
                self.qFeatText1,
                _,
                _,
                _,
            ) = self.get_dataset_embeddings(embed_type=embed_type)

            self.final_dim_size = self.dbFeatImg1.shape[1]

            pca = PCA(n_components=self.final_dim_size)
            pca.fit(self.dbFeatImg1)
            self.dbFeatImg1PCA = pca.transform(self.dbFeatImg1)
            self.qFeatImg1PCA = pca.transform(self.qFeatImg1)

            dbMean = np.mean(self.dbFeatImg1PCA, axis=0)
            dbStd = np.std(self.dbFeatImg1PCA, axis=0)

            self.dbFeatImg1PCA = (self.dbFeatImg1PCA - dbMean) / dbStd
            self.qFeatImg1PCA = (self.qFeatImg1PCA - dbMean) / dbStd


            # calculate this only when sim_space is not present
            if not os.path.exists('data/pkl/' + f"{self.__class__.__name__}_{split}_teacher_sim_space_{embed_type}.npy"):
                # Calculate cosine similarity one row at a time
                print('Calculating Teacher Sim Space')
                self.teacher_sim_space = torch.zeros((self.dbFeatImg1PCA.shape[0], self.dbFeatImg1PCA.shape[0]))
                for i in tqdm(range(self.teacher_sim_space.shape[0])):
                    self.teacher_sim_space[i] = F.cosine_similarity(torch.tensor(self.dbFeatImg1PCA[i]).unsqueeze(0), torch.tensor(self.dbFeatImg1PCA), dim=1)

                # save teacher_sim_space
                np.save('data/pkl/' + f"{self.__class__.__name__}_{split}_teacher_sim_space_{embed_type}.npy", self.teacher_sim_space.numpy())
            else:
                self.teacher_sim_space = torch.tensor(np.load('data/pkl/' + f"{self.__class__.__name__}_{split}_teacher_sim_space_{embed_type}.npy"))

            self.teacher_topk_indices = self.teacher_sim_space.topk(dim=1, k=500)[1]
            self.teacher_topk_values = self.teacher_sim_space.topk(dim=1, k=500)[0]

    
    def get_metadata(self):
        metric_dataset = self.metric_dataset
        self.all_img_files = np.array(metric_dataset.image_paths)
        self.dbFiles = []
        self.qFiles = []
        dbFileMap = []
        qFileMap = []

        gt = []
        instances_covers = []
        instance_labels = np.array(metric_dataset.instance_labels)
        for idx in tqdm(range(len(metric_dataset))):
            instance = instance_labels[idx]
            
            if instance in instances_covers:
                continue

            instance_idxs = np.where(np.array(instance_labels) == instance)[0]
            if len(instance_idxs) > 2:
                self.qFiles.extend(self.all_img_files[instance_idxs[:2]])

                qFileMap.extend(list(instance_idxs[:2]))

                gt.append(list(np.arange(len(instance_idxs[2:])) + len(self.dbFiles)))
                gt.append(list(np.arange(len(instance_idxs[2:])) + len(self.dbFiles)))
                self.dbFiles.extend(self.all_img_files[instance_idxs[2:]])
                dbFileMap.extend(list(instance_idxs[2:]))
            
                instances_covers.append(instance)

        self.all_img_files = self.dbFiles + self.qFiles
        self.file_map = dbFileMap + qFileMap
        return self.all_img_files, gt, len(qFileMap), len(dbFileMap)

    def get_dataset_embeddings(self, embed_type="clip"):
        image_emb_file = (
            self.dir + f"image_embeddings_sop_{embed_type}_{self.split}.npy"
        )

        emb_function = self.embedding_function_mapping[embed_type]


        all_img_files = self.all_img_files

        # Get embeddings will return normalized embeddings
        if not os.path.exists(image_emb_file):
            img_emb = emb_function(self.metric_dataset.image_paths) # TODO: Change this
            np.save(image_emb_file, img_emb)
        else:
            img_emb = np.load(image_emb_file)

        img_emb = img_emb[self.file_map] 

        text_emg = None
        dbFiles = all_img_files
        qFiles = all_img_files

        self.dbFeatImg = img_emb.astype("float32")[:self.numDb]
        self.qFeatImg = img_emb.astype("float32")[self.numDb:]
        self.dbFeatText = None
        self.qFeatText = None

        return (
            self.dbFeatImg,
            self.qFeatImg,
            self.dbFeatText,
            self.qFeatText,
            self.gt,
            self.numDb,
            self.numQ,
        )
    
    def __getitem__(self, index) -> Any:
        indices = self.teacher_topk_indices[index]
        img_emb1 = torch.tensor(self.dbFeatImg1PCA[indices])
        values = self.teacher_topk_values[index]

        return img_emb1, values

    
    def __len__(self) -> int:
        return self.numDb

class DatasetCUB(PRDataset):
    def __init__(self, split="test"):
        self.dir = '/store/ /datasets/CUB_200_2011/CUB_200_2011/'
        self.text_desc_file = self.dir + "text_descriptions_CUB.pkl"
        self.metric_dataset = Cub200(train = True if split == "train" else False)


        _, self.gt, self.numQ, self.numDb = self.get_metadata()
        self.embedding_function_mapping = get_embedding_function_mapping()
        self.split = split


    # def get_metadata(self):
    #     metric_dataset = self.metric_dataset
    #     self.all_img_files = metric_dataset.image_paths
    #     gt = []
    #     instance_labels = np.array(metric_dataset.instance_labels)
    #     for idx in tqdm(range(len(metric_dataset))):
    #         instance = instance_labels[idx]
    #         gtelem = np.where(np.array(instance_labels) == instance)[0]
    #         gtelem = np.delete(gtelem, np.where(gtelem == idx)[0])
    #         gt.append(gtelem)

    #     return self.all_img_files, gt, len(self.all_img_files), len(self.all_img_files)
    
    def get_metadata(self):
        metric_dataset = self.metric_dataset
        self.all_img_files = np.array(metric_dataset.image_paths)
        self.dbFiles = []
        self.qFiles = []
        dbFileMap = []
        qFileMap = []

        gt = []
        instances_covers = []
        instance_labels = np.array(metric_dataset.instance_labels)
        for idx in tqdm(range(len(metric_dataset))):
            instance = instance_labels[idx]
            
            if instance in instances_covers:
                continue

            instance_idxs = np.where(np.array(instance_labels) == instance)[0]
            if len(instance_idxs) > 12:
                num_query = 10
                self.qFiles.extend(self.all_img_files[instance_idxs[:num_query]])

                qFileMap.extend(list(instance_idxs[:num_query]))

                for qidx in range(num_query):
                    gt.append(list(np.arange(len(instance_idxs[num_query:])) + len(self.dbFiles)))
                self.dbFiles.extend(self.all_img_files[instance_idxs[num_query:]])
                dbFileMap.extend(list(instance_idxs[num_query:]))
            
                instances_covers.append(instance)

        self.all_img_files = self.dbFiles + self.qFiles
        self.file_map = dbFileMap + qFileMap
        return self.all_img_files, gt, len(qFileMap), len(dbFileMap)

    def get_dataset_embeddings(self, embed_type="clip"):
        image_emb_file = (
            self.dir + f"image_embeddings_cub_{embed_type}_{self.split}.npy"
        )

        emb_function = self.embedding_function_mapping[embed_type]


        all_img_files = self.all_img_files

        # Get embeddings will return normalized embeddings
        if not os.path.exists(image_emb_file):
            img_emb = emb_function(self.metric_dataset.image_paths) # TODO: Change this
            np.save(image_emb_file, img_emb)
        else:
            img_emb = np.load(image_emb_file)

        img_emb = img_emb[self.file_map] 

        text_emg = None
        dbFiles = all_img_files
        qFiles = all_img_files

        self.dbFeatImg = img_emb.astype("float32")[:self.numDb]
        self.qFeatImg = img_emb.astype("float32")[self.numDb:]
        self.dbFeatText = None
        self.qFeatText = None

        return (
            self.dbFeatImg,
            self.qFeatImg,
            self.dbFeatText,
            self.qFeatText,
            self.gt,
            self.numDb,
            self.numQ,
        )

class DatasetWaterBird(PRDataset):
    def __init__(self):
        self.dir = "/nas/tirr/waterbird/dataset/waterbird_landfraction50_forest2water2/"
        self.text_desc_file = self.dir + "text_descriptions_waterbird.pkl"
        train_img_folder = self.dir + 'train/real/'
        train_img_files = [train_img_folder + f for f in os.listdir(train_img_folder) if f.endswith('.jpg')]
        DB_metadata_csv = pandas.read_csv(self.dir + 'train_real_metadata.csv')
        DB_metadata_csv['img_filename_no_path'] = DB_metadata_csv['img_filename'].apply(lambda x: x.split('/')[-1].replace(".jpg", ""))
        gt_DB = []
        for img_name in train_img_files:
            img_name = img_name.replace(".jpg", "").split('/')[-1]
            gt_DB.append(DB_metadata_csv[DB_metadata_csv['img_filename_no_path'] == img_name]['y'].values[0])
        test_img_folder = self.dir + 'test/'
        test_img_files = [test_img_folder + f for f in os.listdir(test_img_folder) if f.endswith('.jpg')]
        Q_metadata_csv = pandas.read_csv(self.dir + 'test_metadata.csv')
        Q_metadata_csv['img_filename_no_path'] = Q_metadata_csv['img_filename'].apply(lambda x: x.split('/')[-1].replace(".jpg", ""))
        for img_name in test_img_files:
            img_name = img_name.replace(".jpg", "").split('/')[-1]
            gt_DB.append(Q_metadata_csv[Q_metadata_csv['img_filename_no_path'] == img_name]['y'].values[0])
        all_img_files = train_img_files + test_img_files

        self.all_img_files = np.array(all_img_files)
        self.gt = np.array(gt_DB)

    def get_dataset_embeddings(self):
        return None

class DatasetInShop(PRDataset):
    def __init__(self, no_training=True, embed_type="dino"):
        '''
        s2t: student to teacher regression
        '''
        self.dir = "/store/ /datasets/InShop/"
        self.text_desc_file = self.dir + "text_descriptions_InShop.pkl"

        self.all_img_files, self.gt, self.numQ, self.numDb = self.get_metadata()
        self.embedding_function_mapping = get_embedding_function_mapping()
        
        if not no_training:
            self.emb_type = embed_type
            # adding suffix 1 here because get_dataset_embeddings also creates a self.dbFeatImg
            (
                self.dbFeatImg1,
                self.qFeatImg1,
                self.dbFeatText1,
                self.qFeatText1,
                _,
                _,
                _,
            ) = self.get_dataset_embeddings(embed_type=embed_type)

            self.final_dim_size = self.dbFeatImg1.shape[1]

            pca = PCA(n_components=self.final_dim_size)
            pca.fit(self.dbFeatImg1)
            self.dbFeatImg1PCA = pca.transform(self.dbFeatImg1)
            self.qFeatImg1PCA = pca.transform(self.qFeatImg1)

            dbMean = np.mean(self.dbFeatImg1PCA, axis=0)
            dbStd = np.std(self.dbFeatImg1PCA, axis=0)

            self.dbFeatImg1PCA = (self.dbFeatImg1PCA - dbMean) / dbStd
            self.qFeatImg1PCA = (self.qFeatImg1PCA - dbMean) / dbStd


            # calculate this only when sim_space is not present
            if not os.path.exists('data/pkl/' + f"inshop_teacher_sim_space_{embed_type}.npy"):
                # Calculate cosine similarity one row at a time
                print('Calculating Teacher Sim Space')
                self.teacher_sim_space = torch.zeros((self.dbFeatImg1PCA.shape[0], self.dbFeatImg1PCA.shape[0]))
                for i in tqdm(range(self.teacher_sim_space.shape[0])):
                    self.teacher_sim_space[i] = F.cosine_similarity(torch.tensor(self.dbFeatImg1PCA[i]).unsqueeze(0), torch.tensor(self.dbFeatImg1PCA), dim=1)

                # save teacher_sim_space
                np.save('data/pkl/' + f"inshop_teacher_sim_space_{embed_type}.npy", self.teacher_sim_space.numpy())
            else:
                self.teacher_sim_space = torch.tensor(np.load('data/pkl/' + f"inshop_teacher_sim_space_{embed_type}.npy"))

            self.teacher_topk_indices = self.teacher_sim_space.topk(dim=1, k=500)[1]
            self.teacher_topk_values = self.teacher_sim_space.topk(dim=1, k=500)[0]



    def get_metadata(self):
        metric_datasetq = InShop( train=False, query=True)
        metric_datasetdb = InShop( train=False, query=False)
        self.all_img_files =  metric_datasetdb.image_paths + metric_datasetq.image_paths
        numQ = len(metric_datasetq)
        numDb = len(metric_datasetdb)

        db_labels = np.array(metric_datasetdb.instance_labels)
        gt = []
        for idx in range(numQ):
            q_label = metric_datasetq.instance_labels[idx]
            gt.append(np.where(db_labels == q_label)[0])
        
        return self.all_img_files, gt, numQ, numDb

    def get_dataset_embeddings(self, embed_type="clip"):
        image_emb_file = (
            self.dir + f"image_embeddings_inshop_{embed_type}.npy"
        )

        emb_function = self.embedding_function_mapping[embed_type]


        all_img_files = self.all_img_files

        # Get embeddings will return normalized embeddings
        if not os.path.exists(image_emb_file):
            img_emb = emb_function(all_img_files) # TODO: Change this
            np.save(image_emb_file, img_emb)
        else:
            img_emb = np.load(image_emb_file)

        text_emg = None
        dbFiles = all_img_files[:self.numDb]
        qFiles = all_img_files[self.numDb:]

        self.dbFeatImg = img_emb.astype("float32")[:self.numDb]
        self.qFeatImg = img_emb.astype("float32")[self.numDb:]
        self.dbFeatText = None
        self.qFeatText = None

        return (
            self.dbFeatImg,
            self.qFeatImg,
            self.dbFeatText,
            self.qFeatText,
            self.gt,
            self.numDb,
            self.numQ,
        )

    def __getitem__(self, index) -> Any:
        indices = self.teacher_topk_indices[index]
        img_emb1 = torch.tensor(self.dbFeatImg1PCA[indices])
        values = self.teacher_topk_values[index]

        return img_emb1, values

    
    def __len__(self) -> int:
        return self.numDb
    

def get_dataset_mapping():
    return {
        "pitts_train": DatasetPittsTrain,
        "kitti": DatasetKITTI,
        "googlelm_retrival": DatasetGoogleLMRetrival,
        "tokyo_val": DatasetTokyoVal,
        "cub": DatasetCUB,
        "waterbird": DatasetWaterBird,
        "inshop": DatasetInShop,
        "sop": DatasetSOP,
        "pitts": DatasetPitts,
    }

if __name__ == "__main__":
    dataset = DatasetWaterBird()
    breakpoint()
