import torch
from multiprocessing import Pool
import multiprocessing as mp
from tirr.utils import singletons

from llava.constants import (
    IMAGE_TOKEN_INDEX,
    DEFAULT_IMAGE_TOKEN,
)
from llava.conversation import conv_templates
from llava.model.builder import load_pretrained_model
from llava.utils import disable_torch_init
from llava.mm_utils import (
    tokenizer_image_token,
    get_model_name_from_path,
    process_images,
)

import os
import pandas
import requests
from PIL import Image
from io import BytesIO
from transformers import TextStreamer
import time
from tqdm import tqdm
from tirr.packages.netvlad import pittsburgh
from tirr.utils.datasets import DatasetKITTI, DatasetGoogleLMRetrival, DatasetCUB, DatasetTokyoVal
import pickle

cfg = singletons.get_cfg()
logger = singletons.get_logger()


def load_image(image_file):
    if image_file.startswith("http") or image_file.startswith("https"):
        response = requests.get(image_file)
        image = Image.open(BytesIO(response.content)).convert("RGB")
    else:
        image = Image.open(image_file).convert("RGB")
    return image

def get_text_descriptions(input_data):
    args, all_img_files = input_data
    # Model
    disable_torch_init()

    model_name = get_model_name_from_path(args.model_path)
    tokenizer, model, image_processor, context_len = load_pretrained_model(
        args.model_path,
        args.model_base,
        model_name,
        args.load_8bit,
        args.load_4bit,
    )

    if "llama-2" in model_name.lower():
        conv_mode = "llava_llama_2"
    elif "mistral" in model_name.lower():
        conv_mode = "mistral_instruct"
    elif "v1.6-34b" in model_name.lower():
        conv_mode = "chatml_direct"
    elif "v1" in model_name.lower():
        conv_mode = "llava_v1"
    elif "mpt" in model_name.lower():
        conv_mode = "mpt"
    else:
        conv_mode = "llava_v0"

    if args.conv_mode is not None and conv_mode != args.conv_mode:
        print(
            "[WARNING] the auto inferred conversation mode is {}, while `--conv-mode` is {}, using {}".format(
                conv_mode, args.conv_mode, args.conv_mode
            )
        )
    else:
        args.conv_mode = conv_mode

    text_descriptions = []

    ### adapt to newest llava 1.2.1
    count = 0
    for image_file in tqdm(all_img_files):
        count += 1
        conv = conv_templates[args.conv_mode].copy()
        # inp = "Describe the static objects in the image like buildings. Do not include dynamic objects such as people, cars, or animals."
        # inp = "Describe the animal in the image. Do not include the background."
        inp = "Describe only the color, shape, form, etc. of the bird. Do not describe anything in the background."
        inp = DEFAULT_IMAGE_TOKEN + "\n" + inp
        conv.append_message(conv.roles[0], inp)
        conv.append_message(conv.roles[1], None)
        prompt = conv.get_prompt()

        images = [load_image(image_file)]
        image_sizes = [x.size for x in images]
        images_tensor = process_images(
            images,
            image_processor,
            model.config
        ).to(model.device, dtype=torch.float16)
        input_ids = (
            tokenizer_image_token(
                prompt, tokenizer, IMAGE_TOKEN_INDEX, return_tensors="pt"
            )
            .unsqueeze(0)
            .cuda()
        )
        with torch.inference_mode():
            output_ids = model.generate(
                input_ids,
                images=images_tensor,
                image_sizes=image_sizes,
                do_sample=True if args.temperature > 0 else False,
                temperature=args.temperature,
                max_new_tokens=1024,
                use_cache=True,
            )

        outputs = tokenizer.batch_decode(output_ids, skip_special_tokens=True)[0].strip()
        text_descriptions.append([image_file, outputs])

    return text_descriptions


def captions_17P(args):
    mp.set_start_method("spawn")
    num_processes = args.num_processes
    p = Pool(processes=num_processes)
    # Get image file name for equal spread data
    dir_17P = "/store/ /datasets/17places/"
    ref_img_files = [
        f'{dir_17P + "ref/" + str(i)}.jpg' for i in range(0, 406)
    ]  # ref images
    query_img_files = [
        f'{dir_17P + "query/" + str(i)}.jpg' for i in range(0, 406)
    ]  # query images
    all_img_files = ref_img_files + query_img_files

    # data = p.map(get_text_descriptions, [(args, all_img_files[]) for i in range(num_processes)])
    # multi-process and divide all_img files into num_processes such that first process gets first 1/num_processes of all_img_files
    data = p.map(
        get_text_descriptions,
        [
            (
                args,
                all_img_files[
                    int(i * len(all_img_files) / num_processes) : int(
                        (i + 1) * len(all_img_files) / num_processes
                    )
                ],
            )
            for i in range(num_processes)
        ],
    )
    complete_data = []
    for d in data:
        complete_data += d
    # Save text_descriptions pickle
    with open(
        "/store/ /datasets/17places/text_descriptions_17P.pkl", "wb"
    ) as f:
        pickle.dump(complete_data, f)
    p.close()


def captions_pittsburgh(args):
    mp.set_start_method("spawn")
    num_processes = args.num_processes
    p = Pool(processes=num_processes)

    dataset = pittsburgh.get_whole_training_set()
    # dataset = pittsburgh.get_whole_val_set()
    all_img_files = dataset.images

    data = p.map(
        get_text_descriptions,
        [
            (
                args,
                all_img_files[
                    int(i * len(all_img_files) / num_processes) : int(
                        (i + 1) * len(all_img_files) / num_processes
                    )
                ],
            )
            for i in range(num_processes)
        ],
    )
    complete_data = data[0].copy()

    for d in data[1:]:
        complete_data += d

    # Save text_descriptions pickle
    with open(
        "/store/ /datasets/pitts250k/text_descriptions_pitts30k_train.pkl",
        "wb",
    ) as f:
        pickle.dump(complete_data, f)

    breakpoint()


def captions_google_lm(args):
    dataset = DatasetGoogleLMRetrival()
    all_img_files, gt, numQ, numDB = dataset.get_metadata()

    print("numDB:", numDB)
    print("numQ:", numQ)

    mp.set_start_method("spawn")
    num_processes = args.num_processes
    p = Pool(processes=num_processes)

    data = p.map(
        get_text_descriptions,
        [
            (
                args,
                all_img_files[
                    int(i * len(all_img_files) / num_processes) : int(
                        (i + 1) * len(all_img_files) / num_processes
                    )
                ],
            )
            for i in range(num_processes)
        ],
    )
    complete_data = data[0].copy()

    for d in data[1:]:
        complete_data += d

    # Save text_descriptions pickle
    with open(
        dataset.text_desc_file,
        "wb",
    ) as f:
        pickle.dump(complete_data, f)

    breakpoint()


def captions_kitti(args):
    dataset = DatasetKITTI()
    all_img_files, gt, numQ, numDB = dataset.get_metadata()

    print("numDB:", numDB)
    print("numQ:", numQ)

    mp.set_start_method("spawn")
    num_processes = args.num_processes
    p = Pool(processes=num_processes)

    data = p.map(
        get_text_descriptions,
        [
            (
                args,
                all_img_files[
                    int(i * len(all_img_files) / num_processes) : int(
                        (i + 1) * len(all_img_files) / num_processes
                    )
                ],
            )
            for i in range(num_processes)
        ],
    )
    complete_data = data[0].copy()

    for d in data[1:]:
        complete_data += d

    # Save text_descriptions pickle
    with open(
        dataset.dir + "text_descriptions_kitti.pkl",
        "wb",
    ) as f:
        pickle.dump(complete_data, f)

    breakpoint()


def captions_CUB(args):
    dataset = DatasetCUB()
    all_img_files, gt, numQ, numDB = dataset.get_metadata()

    print("numDB:", numDB)
    print("numQ:", numQ)

    mp.set_start_method("spawn")
    num_processes = args.num_processes
    p = Pool(processes=num_processes)

    data = p.map(
        get_text_descriptions,
        [
            (
                args,
                all_img_files[
                    int(i * len(all_img_files) / num_processes) : int(
                        (i + 1) * len(all_img_files) / num_processes
                    )
                ],
            )
            for i in range(num_processes)
        ],
    )
    complete_data = data[0].copy()

    for d in data[1:]:
        complete_data += d

    # Save text_descriptions pickle
    with open(
        dataset.text_desc_file,
        "wb",
    ) as f:
        pickle.dump(complete_data, f)

    breakpoint()


def captions_tokyo(args):
    dataset = DatasetTokyoVal()
    all_img_files, gt, numQ, numDB = dataset.get_metadata()

    print("numDB:", numDB)
    print("numQ:", numQ)

    num_processes = args.num_processes
    p = Pool(processes=num_processes)

    data = p.map(
        get_text_descriptions,
        [
            (
                args,
                all_img_files[
                    int(i * len(all_img_files) / num_processes) : int(
                        (i + 1) * len(all_img_files) / num_processes
                    )
                ],
            )
            for i in range(num_processes)
        ],
    )
    complete_data = data[0].copy()

    for d in data[1:]:
        complete_data += d

    # Save text_descriptions pickle
    with open(
        dataset.text_desc_file,
        "wb",
    ) as f:
        pickle.dump(complete_data, f)

    breakpoint()


def captions_waterbird(args):
    root_dir = "/nas/tirr/waterbird/dataset/waterbird_landfraction50_forest2water2/"
    train_img_folder = root_dir + 'train/real/'
    train_img_files = [train_img_folder + f for f in os.listdir(train_img_folder) if f.endswith('.jpg')]
    DB_metadata_csv = pandas.read_csv(root_dir + 'train_real_metadata.csv')
    DB_metadata_csv['img_filename_no_path'] = DB_metadata_csv['img_filename'].apply(lambda x: x.split('/')[-1].replace(".jpg", ""))
    gt_DB = []
    for img_name in train_img_files:
        img_name = img_name.replace(".jpg", "").split('/')[-1]
        gt_DB.append(DB_metadata_csv[DB_metadata_csv['img_filename_no_path'] == img_name]['y'].values[0])
    test_img_folder = root_dir + 'test/'
    test_img_files = [test_img_folder + f for f in os.listdir(test_img_folder) if f.endswith('.jpg')]
    Q_metadata_csv = pandas.read_csv(root_dir + 'test_metadata.csv')
    Q_metadata_csv['img_filename_no_path'] = Q_metadata_csv['img_filename'].apply(lambda x: x.split('/')[-1].replace(".jpg", ""))
    for img_name in test_img_files:
        img_name = img_name.replace(".jpg", "").split('/')[-1]
        gt_DB.append(Q_metadata_csv[Q_metadata_csv['img_filename_no_path'] == img_name]['y'].values[0])
    all_img_files = train_img_files + test_img_files
    numDB = len(all_img_files)

    val_img_folder = root_dir + 'val/'
    val_img_files = [val_img_folder + f for f in os.listdir(val_img_folder) if f.endswith('.jpg')]
    gt_val = []
    val_metadata_csv = pandas.read_csv(root_dir + 'val_metadata.csv')
    val_metadata_csv['img_filename_no_path'] = val_metadata_csv['img_filename'].apply(lambda x: x.split('/')[-1].replace(".jpg", ""))
    for img_name in val_img_files:
        img_name = img_name.replace(".jpg", "").split('/')[-1]
        gt_val.append(val_metadata_csv[val_metadata_csv['img_filename_no_path'] == img_name]['y'].values[0])
    numQ = len(val_img_files)

    print("numDB:", numDB)
    print("numQ:", numQ)

    try:
        torch.multiprocessing.set_start_method('spawn', force=True)
        mp.set_start_method("spawn", force=True)
        num_processes = args.num_processes
        p = Pool(processes=num_processes)
        print("num_processes:", num_processes)

        data = p.map(
            get_text_descriptions,
            [
                (
                    args,
                    all_img_files[
                        int(i * len(all_img_files) / num_processes) : int(
                            (i + 1) * len(all_img_files) / num_processes
                        )
                    ],
                )
                for i in range(num_processes)
            ],
        )
        all_data = data[0].copy()
    except RuntimeError:
        print("-----------------------------------------------")
        print("RuntimeError.")
        all_data = get_text_descriptions((args, all_img_files))

    # Save text_descriptions pickle
    with open(
        root_dir + "text_descriptions_waterbird_train_test.pkl",
        "wb",
    ) as f:
        pickle.dump(all_data, f)

    val_data = get_text_descriptions((args, val_img_files))
    with open(
        root_dir + "text_descriptions_waterbird_val.pkl",
        "wb",
    ) as f:
        pickle.dump(val_data, f)

    breakpoint()


if __name__ == "__main__":
    mp.set_start_method("spawn")

    # captions_kitti(cfg.llava)
    # captions_google_lm(cfg.llava)
    # captions_CUB(cfg.llava)
    # captions_tokyo(cfg.llava)
    captions_waterbird(cfg.llava)

"""
CUDA_VISIBLE_DEVICES=0 python generate_captions.py

LLAVA Dependency:
llava 1.1.0 requires timm==0.6.13, but you have timm 0.9.8 which is incompatible.                     
llava 1.1.0 requires transformers==4.31.0, but you have transformers 4.38.0.dev0 which is incompatible.
"""
