import open_clip
import numpy as np
from PIL import Image
import torch
from tqdm import tqdm
import time
from tirr.packages.netvlad import nvmodel
from tirr.packages.netvlad.pittsburgh import input_transform as nv_nput_transform
from transformers import AutoTokenizer, AutoModelForSequenceClassification, AutoImageProcessor, AutoModel
from sentence_transformers import SentenceTransformer, util
from transformers import ViTImageProcessor, ViTModel
from tirr.nn.models import CLIPCompressorIT
from torchvision import transforms


# clip_imgs in batch with gpu
def clip_imgs(img_files : list, batch_size : int = 32):
    # model, _, preprocess = open_clip.create_model_and_transforms('ViT-bigG-14', pretrained='laion2B-39B-b160k')
    # tokenizer = open_clip.get_tokenizer('ViT-bigG-14')
    # model, _, preprocess= open_clip.create_model_and_transforms('hf-hub:laion/CLIP-ViT-bigG-14-laion2B-39B-b160k')
    # tokenizer = open_clip.get_tokenizer('hf-hub:laion/CLIP-ViT-bigG-14-laion2B-39B-b160k')
    model, preprocess = open_clip.create_model_from_pretrained('hf-hub:timm/ViT-B-16-SigLIP')
    tokenizer = open_clip.get_tokenizer('hf-hub:timm/ViT-B-16-SigLIP')
    model = model.cuda()
    img_embeddings = []
    with torch.no_grad(), torch.cuda.amp.autocast():
        for i in tqdm(range(0, len(img_files), batch_size)):
            batch = []
            for img_file in img_files[i:i+batch_size]:
                image = preprocess(Image.open(img_file)).unsqueeze(0)
                batch.append(image)
            batch = torch.cat(batch, dim=0)
            batch = batch.cuda()
            image_features = model.encode_image(batch)
            image_features /= image_features.norm(dim=-1, keepdim=True)
            img_embeddings.append(image_features.detach().cpu().numpy())
    img_embeddings = np.concatenate(img_embeddings, axis=0)

    return img_embeddings


# clip text in batch with gpu
def clip_text(text, batch_size : int = 32):
    assert batch_size ==1, "Batch size for text should be 1. Due to the unknown tokenizer bug in open_clip."
    model, _, preprocess= open_clip.create_model_and_transforms('hf-hub:laion/CLIP-ViT-bigG-14-laion2B-39B-b160k')
    tokenizer = open_clip.get_tokenizer('hf-hub:laion/CLIP-ViT-bigG-14-laion2B-39B-b160k')
    model = model.cuda()

    text_features = []
    with torch.no_grad(), torch.cuda.amp.autocast():
        for i in tqdm(range(0, len(text), batch_size)):
            batch = text[i:i+batch_size][0]
            batch = tokenizer(batch)
            batch = batch.cuda()
            batch = model.encode_text(batch)
            batch /= batch.norm(dim=-1, keepdim=True)
            text_features.append(batch.detach().cpu().numpy())
            assert len(text_features) == i+1
    text_features = np.concatenate(text_features, axis=0)

    return text_features

def gtr_text(text, batch_size : int = 32):
    model = SentenceTransformer('sentence-transformers/gtr-t5-large')
    model = model.cuda()
    text_features = model.encode(text)
    return text_features


# def bert(text, batch_size : int = 4):
#     model = AutoModelForSequenceClassification.from_pretrained(
#         "bert-base-uncased",
#         trust_remote_code=True
#     )

#     tokenizer = AutoTokenizer.from_pretrained(
#         "bert-base-uncased",
#     )

#     model = model.cuda()
#     text_features = []
#     with torch.no_grad():
#         for i in tqdm(range(0, len(text), batch_size)):
#             batch = text[i:i+batch_size]
#             batch = tokenizer(
#                 batch,
#                 return_tensors="pt",
#                 padding="max_length",
#                 return_token_type_ids=False,
#                 truncation=True,
#             )
#             batch = batch.to("cuda")
#             outputs = model(**batch)
#             embeddings = outputs['sentence_embedding']
#             embeddings /= embeddings.norm(dim=-1, keepdim=True)
#             text_features.append(embeddings.detach().cpu().numpy())

#     return text_features

def bert(text, batch_size : int = 4):
    model = SentenceTransformer('sentence-transformers/all-mpnet-base-v2')
    model = model.cuda()
    text_features = model.encode(text)
    text_features = text_features / np.linalg.norm(text_features, axis=1, keepdims=True)
    
    # with torch.no_grad():
    #     for i in tqdm(range(0, len(text), batch_size)):
    #         batch = text[i:i+batch_size]
    #         batch = model.encode(batch, convert_to_tensor=True)
    #         text_features.append(batch.detach().cpu().numpy())

    return text_features

def netvlad(img_files : list, batch_size : int = 32):
    model = nvmodel.get_model()
    model = model.cuda()
    img_embeddings = []
    transform = nv_nput_transform()
    with torch.no_grad():
        for i in tqdm(range(0, len(img_files), batch_size)):
            batch = []
            for img_file in img_files[i:i+batch_size]:
                image = Image.open(img_file)
                image = transform(image)
                image = image.unsqueeze(0)
                batch.append(image)
            batch = torch.cat(batch, dim=0)
            batch = batch.cuda()
            image_features = model.encoder(batch)
            image_features = model.pool(image_features)
            img_embeddings.append(image_features.detach().cpu().numpy())
    img_embeddings = np.concatenate(img_embeddings, axis=0)

    return img_embeddings


def dino_vitb8(img_files : list, batch_size : int = 32):
    '''
    Number of params: 86.3M
    total_params_in_M = sum(p.numel() for p in model.parameters()) / 1_000_000
    '''
    print("Loading DINO")
    processor = ViTImageProcessor.from_pretrained('facebook/dino-vitb8')
    model = ViTModel.from_pretrained('facebook/dino-vitb8')
    model = model.cuda()
    img_embeddings = []
    with torch.no_grad():
        for i in tqdm(range(0, len(img_files), batch_size)):
            images = []
            for img_file in img_files[i:i+batch_size]:
                image = Image.open(img_file).convert('RGB')
                images.append(image)
            batch = processor(images, return_tensors="pt").to("cuda")
            outputs = model(**batch)
            image_features = outputs.last_hidden_state
            image_features = image_features.mean(dim=1)
            img_embeddings.append(image_features.detach().cpu().numpy())

    img_embeddings = np.concatenate(img_embeddings, axis=0)

    return img_embeddings

def dino_vitb16(img_files : list, batch_size : int = 32):
    # pass
    print("Loading DINO vitb16")
    processor = ViTImageProcessor.from_pretrained('facebook/dino-vitb16')
    model = ViTModel.from_pretrained('facebook/dino-vitb16')
    model = model.cuda()
    img_embeddings = []
    with torch.no_grad():
        for i in tqdm(range(0, len(img_files), batch_size)):
            images = []
            for img_file in img_files[i:i+batch_size]:
                image = Image.open(img_file).convert('RGB')
                images.append(image)
            batch = processor(images, return_tensors="pt").to("cuda")
            outputs = model(**batch)
            image_features = outputs.last_hidden_state
            image_features = image_features.mean(dim=1)
            img_embeddings.append(image_features.detach().cpu().numpy())

    img_embeddings = np.concatenate(img_embeddings, axis=0)

    return img_embeddings

def dino_vits8(img_files : list, batch_size : int = 32):
    # pass
    print("Loading DINO")
    processor = ViTImageProcessor.from_pretrained('facebook/dino-vits8')
    model = ViTModel.from_pretrained('facebook/dino-vits8')
    model = model.cuda()
    img_embeddings = []
    with torch.no_grad():
        for i in tqdm(range(0, len(img_files), batch_size)):
            images = []
            for img_file in img_files[i:i+batch_size]:
                image = Image.open(img_file).convert('RGB')
                images.append(image)
            batch = processor(images, return_tensors="pt").to("cuda")
            outputs = model(**batch)
            image_features = outputs.last_hidden_state
            image_features = image_features.mean(dim=1)
            img_embeddings.append(image_features.detach().cpu().numpy())

    img_embeddings = np.concatenate(img_embeddings, axis=0)

    return img_embeddings

def dino_vits16(img_files : list, batch_size : int = 32):
    # pass
    print("Loading DINO")
    processor = ViTImageProcessor.from_pretrained('facebook/dino-vits16')
    model = ViTModel.from_pretrained('facebook/dino-vits16')
    model = model.cuda()
    img_embeddings = []
    with torch.no_grad():
        for i in tqdm(range(0, len(img_files), batch_size)):
            images = []
            for img_file in img_files[i:i+batch_size]:
                image = Image.open(img_file).convert('RGB')
                images.append(image)
            batch = processor(images, return_tensors="pt").to("cuda")
            outputs = model(**batch)
            image_features = outputs.last_hidden_state
            image_features = image_features.mean(dim=1)
            img_embeddings.append(image_features.detach().cpu().numpy())

    img_embeddings = np.concatenate(img_embeddings, axis=0)
    return img_embeddings


def vit(img_files : list, batch_size : int = 32):
    # pass
    print("Loading VIT")
    # processor = ViTImageProcessor.from_pretrained('google/vit-base-patch16-224-in21k')
    # model = ViTModel.from_pretrained('google/vit-base-patch16-224-in21k')
    processor = ViTImageProcessor.from_pretrained('google/vit-large-patch32-384')
    model = ViTModel.from_pretrained('google/vit-large-patch32-384')
    
    
    model = model.cuda()
    img_embeddings = []
    with torch.no_grad():
        for i in tqdm(range(0, len(img_files), batch_size)):
            images = []
            for img_file in img_files[i:i+batch_size]:
                image = Image.open(img_file).convert('RGB')
                images.append(image)
            batch = processor(images, return_tensors="pt").to("cuda")
            outputs = model(**batch)
            image_features = outputs.last_hidden_state
            image_features = image_features.mean(dim=1)
            img_embeddings.append(image_features.detach().cpu().numpy())

    img_embeddings = np.concatenate(img_embeddings, axis=0)

    return img_embeddings

def dinov2_giant(img_files : list, batch_size : int = 32):
    # pass
    print("Loading DINO")
    processor = AutoImageProcessor.from_pretrained('facebook/dinov2-giant')
    model = AutoModel.from_pretrained('facebook/dinov2-giant')
    model = model.cuda()
    img_embeddings = []
    with torch.no_grad():
        for i in tqdm(range(0, len(img_files), batch_size)):
            images = []
            for img_file in img_files[i:i+batch_size]:
                image = Image.open(img_file)
                images.append(image)
            batch = processor(images, return_tensors="pt").to("cuda")
            outputs = model(**batch)
            image_features = outputs.last_hidden_state
            image_features = image_features.mean(dim=1)
            img_embeddings.append(image_features.detach().cpu().numpy())

    img_embeddings = np.concatenate(img_embeddings, axis=0)

    return img_embeddings

def dinov2_large(img_files : list, batch_size : int = 32):
    # pass
    print("Loading DINO")
    '''
    self.dino_model: nn.Module = torch.hub.load(
                'facebookresearch/dinov2', dino_model)
    '''
    processor = AutoImageProcessor.from_pretrained('facebook/dinov2-large')
    model = AutoModel.from_pretrained('facebook/dinov2-large')
    model = model.cuda()
    img_embeddings = []
    with torch.no_grad():
        for i in tqdm(range(0, len(img_files), batch_size)):
            images = []
            for img_file in img_files[i:i+batch_size]:
                image = Image.open(img_file)
                images.append(image)
            batch = processor(images, return_tensors="pt").to("cuda")
            outputs = model(**batch)
            image_features = outputs.last_hidden_state
            image_features = image_features.mean(dim=1)
            img_embeddings.append(image_features.detach().cpu().numpy())

    img_embeddings = np.concatenate(img_embeddings, axis=0)

    return img_embeddings

def dinov2_vitg14(img_files : list, batch_size : int = 32):
     # pass
    print("Loading DINO")
    '''
    self.dino_model: nn.Module = torch.hub.load(
                'facebookresearch/dinov2', dino_model)
    '''
    model = torch.hub.load('facebookresearch/dinov2', 'dinov2_vitg14')
    model.eval()
    model = model.cuda()
    img_embeddings = []
    # Define image transformation
    transform = transforms.Compose([
        transforms.Resize((224, 224)),
        transforms.ToTensor(),
        transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]),
    ])
    with torch.no_grad():
        for i in tqdm(range(0, len(img_files), batch_size)):
            images = []
            for img_file in img_files[i:i+batch_size]:
                image = Image.open(img_file)
                image = transform(image).unsqueeze(0)
                images.append(image)
            batch = torch.cat(images, dim=0).cuda()
            outputs = model(batch)
            img_embeddings.append(outputs.detach().cpu().numpy())

    img_embeddings = np.concatenate(img_embeddings, axis=0)

    return img_embeddings

def dinov2_small(img_files : list, batch_size : int = 32):
    # pass
    print("Loading DINO")
    processor = AutoImageProcessor.from_pretrained('facebook/dinov2-small')
    model = AutoModel.from_pretrained('facebook/dinov2-small')
    model = model.cuda()
    img_embeddings = []
    with torch.no_grad():
        for i in tqdm(range(0, len(img_files), batch_size)):
            images = []
            for img_file in img_files[i:i+batch_size]:
                image = Image.open(img_file)
                images.append(image)
            batch = processor(images, return_tensors="pt").to("cuda")
            outputs = model(**batch)
            image_features = outputs.last_hidden_state
            image_features = image_features.mean(dim=1)
            img_embeddings.append(image_features.detach().cpu().numpy())

    img_embeddings = np.concatenate(img_embeddings, axis=0)

    return img_embeddings

def dinov2_base(img_files : list, batch_size : int = 32):
    # pass
    print("Loading DINO")
    processor = AutoImageProcessor.from_pretrained('facebook/dinov2-base')
    model = AutoModel.from_pretrained('facebook/dinov2-base')
    model = model.cuda()
    img_embeddings = []
    with torch.no_grad():
        for i in tqdm(range(0, len(img_files), batch_size)):
            images = []
            for img_file in img_files[i:i+batch_size]:
                image = Image.open(img_file)
                images.append(image)
            batch = processor(images, return_tensors="pt").to("cuda")
            outputs = model(**batch)
            image_features = outputs.last_hidden_state
            image_features = image_features.mean(dim=1)
            img_embeddings.append(image_features.detach().cpu().numpy())

    img_embeddings = np.concatenate(img_embeddings, axis=0)

    return img_embeddings

def cosplace(img_files : list, batch_size : int = 32):
    transforms_list = []

    # transforms_list += [transforms.Resize(image_size, antialias=True)]
    transforms_list += [
            transforms.ToTensor(),
            transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]),
        ]
    transform = transforms.Compose(transforms_list)
    
    model = torch.hub.load("gmberton/cosplace", "get_trained_model", backbone="ResNet50", fc_output_dim=2048)
    model = model.cuda()

    img_embeddings = []
    with torch.no_grad():
        for i in tqdm(range(0, len(img_files), batch_size)):
            images = []
            for img_file in img_files[i:i+batch_size]:
                image = Image.open(img_file).convert('RGB')
                image = transform(image).unsqueeze(0)
                images.append(image)
            batch = torch.cat(images, dim=0).cuda()
            outputs = model(batch)
            img_embeddings.append(outputs.detach().cpu().numpy())

    img_embeddings = np.concatenate(img_embeddings, axis=0)

    return img_embeddings


def image_text_combine(img_emb, text_emb, batch_size : int = 32, model = None):
    if model is None:
        # load clip compressor
        model = CLIPCompressorIT()
        # load saved model
        model.load_state_dict(torch.load("data/nn/best_model_clip_compressor.pt"))
        model = model.cuda()
        # combine embeddings

    all_emb = []

    with torch.no_grad():
        for i in tqdm(range(0, len(img_emb), batch_size)):
            img_batch = img_emb[i:i+batch_size]
            text_batch = text_emb[i:i+batch_size]

            img_batch = torch.tensor(img_batch).cuda()
            text_batch = torch.tensor(text_batch).cuda()
            img_emb_out, text_emb_out = model(img_batch, text_batch)

            img_emb_out = img_emb_out.detach().cpu().numpy()
            text_emb_out = text_emb_out.detach().cpu().numpy()

            #concatenate 
            combined_emb = np.concatenate([img_emb_out, text_emb_out], axis=1)
            all_emb.append(combined_emb)

    return np.concatenate(all_emb, axis=0)


def get_embedding_function_mapping():
    return {
        "clip": clip_imgs,
        "netvlad": netvlad,
        "dinob8": dino_vitb8,
        "dinob16": dino_vitb16,
        "dinos8": dino_vits8,
        "dinos16": dino_vits16,
        "dinov2_large": dinov2_large,
        "dinov2_small": dinov2_small,
        "dinov2_base": dinov2_base,
        "dinov2_giant": dinov2_giant,
        "dinov2_vitg14": dinov2_vitg14,
        "vit": vit,
        "cosplace": cosplace,
    }


