import numpy as np
from tirr.utils import get_predictions


def mean_average_precision_at_k(gt_retrieval, prediction_retrieval, k):
    num_queries = len(gt_retrieval)
    average_precisions = np.zeros(num_queries)

    for i in range(num_queries):
        num_correct = np.isin(
            prediction_retrieval[i][:k], gt_retrieval[i]
        ).cumsum()
        precision_at_k = num_correct / np.arange(1, k + 1)
        average_precisions[i] = np.sum(
            precision_at_k * (num_correct > 0)
        ) / min(k, len(gt_retrieval[i]))

    return np.mean(average_precisions)


def get_recall(
    qFeat,
    dbFeat,
    gt,
    numQ,
    k=3,
    q_in_db=False,
    num_similar_db=6,
    distance="cosine",
):
    # normalize
    dbFeat /= np.linalg.norm(dbFeat, axis=1, keepdims=True)
    qFeat /= np.linalg.norm(qFeat, axis=1, keepdims=True)

    if q_in_db:
        distances, predictions = get_predictions.faiss_preds(
            qFeat, dbFeat, k + num_similar_db, distance
        )
        predictions = predictions[:, num_similar_db:]
    else:
        distances, predictions = get_predictions.faiss_preds(
            qFeat, dbFeat, k, distance
        )

    correct_preds = 0
    for qIx, predNV in enumerate(predictions):
        if np.any(np.in1d(predNV[:k], gt[qIx])):
            correct_preds += 1
    recall = correct_preds / numQ
    return recall


def get_map(
    qFeat,
    dbFeat,
    gt,
    numQ,
    k=3,
    q_in_db=False,
    num_similar_db=6,
    distance="cosine",
):
    # normalize
    dbFeat /= np.linalg.norm(dbFeat, axis=1, keepdims=True)
    qFeat /= np.linalg.norm(qFeat, axis=1, keepdims=True)

    if q_in_db:
        distances, predictions = get_predictions.faiss_preds(
            qFeat, dbFeat, k + num_similar_db, distance
        )
        predictions = predictions[:, num_similar_db:]
    else:
        distances, predictions = get_predictions.faiss_preds(
            qFeat, dbFeat, k, distance
        )

    map_at_k = mean_average_precision_at_k(gt, predictions, k)
    return map_at_k


def get_recall_custom(
    qFeatImg,
    dbFeatImg,
    qFeatText,
    dbFeatText,
    gt,
    numQ,
    n=3,
    q_in_db=False,
    mean=True,
):
    # normalize
    dbFeatImg /= np.linalg.norm(dbFeatImg, axis=1, keepdims=True)
    qFeatImg /= np.linalg.norm(qFeatImg, axis=1, keepdims=True)
    dbFeatText /= np.linalg.norm(dbFeatText, axis=1, keepdims=True)
    qFeatText /= np.linalg.norm(qFeatText, axis=1, keepdims=True)

    text_filter_num = 5000
    if q_in_db:
        raise NotImplementedError
    else:
        # cosine similarity text using np
        distances_text = np.dot(qFeatText, dbFeatText.T)
        predictions_text = np.argsort(-distances_text, axis=1)

        distances_img = np.dot(qFeatImg, dbFeatImg.T)  # this is similarity
        predictions_img = np.argsort(-distances_img, axis=1)

        distances_comb = distances_img.copy()
        for i in range(predictions_text.shape[0]):
            distances_comb[i, predictions_text[i][-text_filter_num:]] = 0

        predictions_comb = np.argsort(-distances_comb, axis=1)

        correct_preds = 0
        for qIx, predNV in enumerate(predictions_img):
            correct_preds += np.mean(np.in1d(predNV[:n], gt[qIx]))
        recall_img = correct_preds / numQ

        correct_preds = 0
        for qIx, predNV in enumerate(predictions_comb):
            correct_preds += np.mean(np.in1d(predNV[:n], gt[qIx]))
        recall_comb = correct_preds / numQ

    return recall_img, recall_comb


def get_metrics_mapping():
    return {
        "recall": get_recall,
        "map": get_map,
        "recall_custom": get_recall_custom,
    }
