import numpy as np

class CustomPCA:
    def __init__(self, n_components=2):
        self.n_components = n_components
        self.projection_matrix = None
        self.sorted_eigenvalues = None
        self.sorted_eigenvectors = None
        self.explained_variance = None
    
    def fit_transform(self, X):
        '''
        X: (n_samples, n_features)
        '''

        # Centering and Standardization
        self.mean = X.mean(axis=0)
        X = X - self.mean

        # Covariance Matrix
        covariance_matrix = np.cov(X, rowvar=False)

        # Eigenvalues and Eigenvectors
        eigenvalues, eigenvectors = np.linalg.eig(covariance_matrix)
        # take only real part
        eigenvalues = np.real(eigenvalues)
        eigenvectors = np.real(eigenvectors)
        
        order_of_importance = np.argsort(eigenvalues)[::-1] 
        self.sorted_eigenvalues = eigenvalues[order_of_importance]
        self.sorted_eigenvectors = eigenvectors[:,order_of_importance] # sort the columns



        self.explained_variance = self.sorted_eigenvalues / np.sum(self.sorted_eigenvalues)

        # Projection Matrix
        self.projection_matrix = self.sorted_eigenvectors[:,:self.n_components]

        # Projection
        X_pca = X.dot(self.projection_matrix)

        return X_pca
    
    def transform(self, X):
        # Centering and Standardization
        X = X - self.mean

        # Projection
        X_pca = X.dot(self.projection_matrix)

        return X_pca
    
    def inverse_transform(self, X):
        X = X.dot(self.projection_matrix.T)
        X = X + self.mean
        return X
    
    def transform_given_components(self, X, evidx):
        # Centering and Standardization
        X = X - X.mean(axis=0)

        # Projection Matrix
        projection_matrix = self.sorted_eigenvectors[:,evidx]

        # Projection
        X_pca = X.dot(projection_matrix)

        return X_pca



class CustomCPCA:
    def __init__(self, n_components=2, alpha=1):
        self.n_components = n_components
        self.projection_matrix = None
        self.sorted_eigenvalues = None
        self.sorted_eigenvectors = None
        self.explained_variance = None
        self.alpha = alpha
    
    def fit_transform(self, fg, bg):
        '''
        fg: foreground data (n_samples, n_features)
        bg: background data (n_samples, n_features)
        '''

        # Centering and Standardization
        self.fg_mean = fg.mean(axis=0)
        fg = fg - self.fg_mean

        self.bg_mean = bg.mean(axis=0)
        bg = bg - self.bg_mean

        # Covariance Matrix
        covariance_matrix_fg = np.cov(fg, rowvar=False)
        covariance_matrix_bg = np.cov(bg, rowvar=False)

        covariance_matrix = covariance_matrix_fg  - self.alpha * covariance_matrix_bg



        # Eigenvalues and Eigenvectors
        eigenvalues, eigenvectors = np.linalg.eig(covariance_matrix)
        # take only real part
        eigenvalues = np.real(eigenvalues)
        eigenvectors = np.real(eigenvectors)
        
        order_of_importance = np.argsort(eigenvalues)[::-1] 
        self.sorted_eigenvalues = eigenvalues[order_of_importance]
        self.sorted_eigenvectors = eigenvectors[:,order_of_importance] # sort the columns



        self.explained_variance = self.sorted_eigenvalues / np.sum(self.sorted_eigenvalues)

        # Projection Matrix
        self.projection_matrix = self.sorted_eigenvectors[:,:self.n_components]

        # Projection
        fg_pca = fg.dot(self.projection_matrix)

        return fg_pca
    
    def transform_fg(self, fg):
        # Centering and Standardization
        fg = fg - self.fg_mean

        # Projection
        fg_pca = fg.dot(self.projection_matrix)

        return fg_pca
    
    def inverse_transform_fg(self, fg):
        fg = fg.dot(self.projection_matrix.T)
        fg = fg - self.bg_mean
        return fg